# This file was auto generated from a SCHISM namelist file on 2025-01-24.

from typing import List, Optional

from pydantic import Field, field_validator, model_validator

from rompy_schism.namelists.basemodel import NamelistBaseModel


class Marco(NamelistBaseModel):
    idelay: Optional[int] = Field(
        0, description="Switch for 7-day delay in zooplankton predation (0: off, 1: on)"
    )
    ndelay: Optional[int] = Field(
        7,
        description="Number of days for zooplankton predation delay when idelay is active",
    )
    ibgraze: Optional[int] = Field(
        0, description="Switch for bottom grazing function (0: off, 1: on)"
    )
    idapt: Optional[int] = Field(
        0, description="Switch for light adaptation (0: off, 1: on)"
    )
    alpha_corr: Optional[float] = Field(
        1.25, description="Correction factor for light adaptation when idapt is active"
    )
    zeptic: Optional[float] = Field(
        10.0,
        description="Euphotic zone depth parameter for light adaptation when idapt is active",
    )
    iz2graze: Optional[int] = Field(
        1, description="Switch for Z2 grazing on S2, Z1, and DN (0: off, 1: on)"
    )
    iout_cosine: Optional[int] = Field(
        0,
        description="CoSiNE model station output option (0: off, 1-5: various output levels)",
    )
    nspool_cosine: Optional[int] = Field(
        30,
        description="Output interval (number of time steps) for CoSiNE model station output",
    )
    ico2s: Optional[int] = Field(
        0,
        description="Switch for CO2 limitation on phytoplankton growth (0: off, 1: on)",
    )
    ispm: Optional[int] = Field(
        0,
        description="Suspended Particulate Matter (SPM) calculation method (0: constant, 1: spatial varying, 2: SED model)",
    )
    spm0: Optional[float] = Field(
        20.0, description="Constant SPM value used when ispm is 0"
    )
    ised: Optional[int] = Field(
        1, description="Switch for sediment flux model (0: off, 1: on)"
    )

    @field_validator("idelay")
    @classmethod
    def validate_idelay(cls, v):
        if v not in [0, 1]:
            raise ValueError("idelay must be 0 or 1")
        return v

    @field_validator("ndelay")
    @classmethod
    def validate_ndelay(cls, v):
        if v < 0:
            raise ValueError("ndelay must be non-negative")
        return v

    @field_validator("ibgraze")
    @classmethod
    def validate_ibgraze(cls, v):
        if v not in [0, 1]:
            raise ValueError("ibgraze must be 0 or 1")
        return v

    @field_validator("idapt")
    @classmethod
    def validate_idapt(cls, v):
        if v not in [0, 1]:
            raise ValueError("idapt must be 0 or 1")
        return v

    @field_validator("alpha_corr")
    @classmethod
    def validate_alpha_corr(cls, v):
        if v <= 0:
            raise ValueError("alpha_corr must be positive")
        return v

    @field_validator("zeptic")
    @classmethod
    def validate_zeptic(cls, v):
        if v <= 0:
            raise ValueError("zeptic must be positive")
        return v

    @field_validator("iz2graze")
    @classmethod
    def validate_iz2graze(cls, v):
        if v not in [0, 1]:
            raise ValueError("iz2graze must be 0 or 1")
        return v

    @field_validator("iout_cosine")
    @classmethod
    def validate_iout_cosine(cls, v):
        if v not in range(6):
            raise ValueError("iout_cosine must be between 0 and 5")
        return v

    @field_validator("nspool_cosine")
    @classmethod
    def validate_nspool_cosine(cls, v):
        if v <= 0:
            raise ValueError("nspool_cosine must be positive")
        return v

    @field_validator("ico2s")
    @classmethod
    def validate_ico2s(cls, v):
        if v not in [0, 1]:
            raise ValueError("ico2s must be 0 or 1")
        return v

    @field_validator("ispm")
    @classmethod
    def validate_ispm(cls, v):
        if v not in [0, 1, 2]:
            raise ValueError("ispm must be 0, 1, or 2")
        return v

    @field_validator("spm0")
    @classmethod
    def validate_spm0(cls, v):
        if v < 0:
            raise ValueError("spm0 must be non-negative")
        return v

    @field_validator("ised")
    @classmethod
    def validate_ised(cls, v):
        if v not in [0, 1]:
            raise ValueError("ised must be 0 or 1")
        return v


class Core(NamelistBaseModel):
    gmaxs: Optional[list] = Field(
        [2.0, 2.5], description="Maximum growth rates for two phytoplankton types"
    )
    gammas: Optional[list] = Field(
        [0.2, 0.075], description="Mortality rates for two phytoplankton types"
    )
    pis: Optional[list] = Field(
        [1.5, 1.5],
        description="Ammonium inhibition parameters for two phytoplankton types",
    )
    kno3s: Optional[list] = Field(
        [1.0, 3.0],
        description="NO3 half-saturation constants for two phytoplankton types",
    )
    knh4s: Optional[list] = Field(
        [0.15, 0.45],
        description="NH4 half-saturation constants for two phytoplankton types",
    )
    kpo4s: Optional[list] = Field(
        [0.1, 0.1],
        description="PO4 half-saturation constants for two phytoplankton types",
    )
    kco2s: Optional[list] = Field(
        [50.0, 50.0],
        description="CO2 half-saturation constants for two phytoplankton types",
    )
    ksio4: Optional[float] = Field(
        4.5, description="SiO4 half-saturation constant for diatoms"
    )
    kns: Optional[list] = Field(
        [0.0, 0.0],
        description="Nighttime uptake rates of NH4 for two phytoplankton types",
    )
    alphas: Optional[list] = Field(
        [0.1, 0.1],
        description="Initial slopes of P-I curve for two phytoplankton types",
    )
    betas: Optional[list] = Field(
        [0.0, 0.0],
        description="Slopes for photo-inhibition for two phytoplankton types",
    )
    aks: Optional[list] = Field(
        [0.75, 0.03, 0.066],
        description="Light extinction coefficients for phytoplankton and suspended particulate matter",
    )
    betaz: Optional[list] = Field(
        [1.35, 0.4], description="Maximum grazing rates for two zooplankton types"
    )
    alphaz: Optional[list] = Field(
        [0.75, 0.75], description="Assimilation rates for two zooplankton types"
    )
    gammaz: Optional[list] = Field(
        [0.2, 0.2], description="Mortality rates for two zooplankton types"
    )
    kez: Optional[list] = Field(
        [0.2, 0.2], description="Excretion rates for two zooplankton types"
    )
    kgz: Optional[list] = Field(
        [0.5, 0.25],
        description="Reference prey concentrations for grazing for two zooplankton types",
    )
    rhoz: Optional[list] = Field(
        [0.6, 0.3, 0.1], description="Prey preference factors of Z2 on (S2, Z1, DN)"
    )
    ipo4: Optional[int] = Field(
        1, description="Flag to add additional PO4 from biogenic silica dissolution"
    )
    tr: Optional[float] = Field(
        20.0,
        description="Reference temperature for temperature adjustment for CoSiNE sink and source",
    )
    kox: Optional[float] = Field(
        30.0, description="Reference oxygen concentration for oxidation"
    )
    wss2: Optional[float] = Field(
        0.2, description="Settling velocity of S2 (phytoplankton type 2)"
    )
    wsdn: Optional[float] = Field(
        1.0, description="Settling velocity of DN (detrital nitrogen)"
    )
    wsdsi: Optional[float] = Field(
        1.0, description="Settling velocity of DSi (dissolved silica)"
    )
    si2n: Optional[float] = Field(
        1.2, description="Silica to nitrogen conversion coefficient"
    )
    p2n: Optional[float] = Field(
        0.0625, description="Phosphorus to nitrogen conversion coefficient"
    )
    o2no: Optional[float] = Field(
        8.625, description="Oxygen to nitrogen (NO3) conversion coefficient"
    )
    o2nh: Optional[float] = Field(
        6.625, description="Oxygen to nitrogen (NH4) conversion coefficient"
    )
    c2n: Optional[float] = Field(
        7.3, description="Carbon to nitrogen conversion coefficient"
    )
    gamman: Optional[float] = Field(0.07, description="Nitrification coefficient")
    pco2a: Optional[float] = Field(391.63, description="Atmospheric CO2 concentration")
    kmdn: Optional[list] = Field(
        [0.009, 0.075],
        description="Remineralization coefficients for DN (detrital nitrogen)",
    )
    kmdsi: Optional[list] = Field(
        [0.0114, 0.015],
        description="Remineralization coefficients for DSi (dissolved silica)",
    )

    @field_validator("gmaxs")
    @classmethod
    def check_gmaxs(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and x > 0 for x in v)
        ):
            raise ValueError("gmaxs must be a list of 2 positive floats")
        return v

    @field_validator("gammas")
    @classmethod
    def check_gammas(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and 0 <= x <= 1 for x in v)
        ):
            raise ValueError("gammas must be a list of 2 floats between 0 and 1")
        return v

    @field_validator("pis")
    @classmethod
    def check_pis(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and x > 0 for x in v)
        ):
            raise ValueError("pis must be a list of 2 positive floats")
        return v

    @field_validator("kno3s")
    @classmethod
    def check_kno3s(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and x > 0 for x in v)
        ):
            raise ValueError("kno3s must be a list of 2 positive floats")
        return v

    @field_validator("knh4s")
    @classmethod
    def check_knh4s(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and x > 0 for x in v)
        ):
            raise ValueError("knh4s must be a list of 2 positive floats")
        return v

    @field_validator("kpo4s")
    @classmethod
    def check_kpo4s(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and x > 0 for x in v)
        ):
            raise ValueError("kpo4s must be a list of 2 positive floats")
        return v

    @field_validator("kco2s")
    @classmethod
    def check_kco2s(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and x > 0 for x in v)
        ):
            raise ValueError("kco2s must be a list of 2 positive floats")
        return v

    @field_validator("ksio4")
    @classmethod
    def check_ksio4(cls, v):
        if not (isinstance(v, float) and v > 0):
            raise ValueError("ksio4 must be a positive float")
        return v

    @field_validator("kns")
    @classmethod
    def check_kns(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and x >= 0 for x in v)
        ):
            raise ValueError("kns must be a list of 2 non-negative floats")
        return v

    @field_validator("alphas")
    @classmethod
    def check_alphas(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and x > 0 for x in v)
        ):
            raise ValueError("alphas must be a list of 2 positive floats")
        return v

    @field_validator("betas")
    @classmethod
    def check_betas(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and x >= 0 for x in v)
        ):
            raise ValueError("betas must be a list of 2 non-negative floats")
        return v

    @field_validator("aks")
    @classmethod
    def check_aks(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 3
            and all(isinstance(x, float) and x > 0 for x in v)
        ):
            raise ValueError("aks must be a list of 3 positive floats")
        return v

    @field_validator("betaz")
    @classmethod
    def check_betaz(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and x > 0 for x in v)
        ):
            raise ValueError("betaz must be a list of 2 positive floats")
        return v

    @field_validator("alphaz")
    @classmethod
    def check_alphaz(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and 0 < x <= 1 for x in v)
        ):
            raise ValueError("alphaz must be a list of 2 floats between 0 and 1")
        return v

    @field_validator("gammaz")
    @classmethod
    def check_gammaz(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and 0 < x <= 1 for x in v)
        ):
            raise ValueError("gammaz must be a list of 2 floats between 0 and 1")
        return v

    @field_validator("kez")
    @classmethod
    def check_kez(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and 0 < x <= 1 for x in v)
        ):
            raise ValueError("kez must be a list of 2 floats between 0 and 1")
        return v

    @field_validator("kgz")
    @classmethod
    def check_kgz(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and x > 0 for x in v)
        ):
            raise ValueError("kgz must be a list of 2 positive floats")
        return v

    @field_validator("rhoz")
    @classmethod
    def check_rhoz(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 3
            and all(isinstance(x, float) and 0 <= x <= 1 for x in v)
        ):
            raise ValueError("rhoz must be a list of 3 floats between 0 and 1")
        return v

    @field_validator("ipo4")
    @classmethod
    def check_ipo4(cls, v):
        if v not in [0, 1]:
            raise ValueError("ipo4 must be either 0 or 1")
        return v

    @field_validator("tr")
    @classmethod
    def check_tr(cls, v):
        if not (isinstance(v, float) and v > 0):
            raise ValueError("tr must be a positive float")
        return v

    @field_validator("kox")
    @classmethod
    def check_kox(cls, v):
        if not (isinstance(v, float) and v > 0):
            raise ValueError("kox must be a positive float")
        return v

    @field_validator("wss2")
    @classmethod
    def check_wss2(cls, v):
        if not (isinstance(v, float) and v >= 0):
            raise ValueError("wss2 must be a non-negative float")
        return v

    @field_validator("wsdn")
    @classmethod
    def check_wsdn(cls, v):
        if not (isinstance(v, float) and v >= 0):
            raise ValueError("wsdn must be a non-negative float")
        return v

    @field_validator("wsdsi")
    @classmethod
    def check_wsdsi(cls, v):
        if not (isinstance(v, float) and v >= 0):
            raise ValueError("wsdsi must be a non-negative float")
        return v

    @field_validator("si2n")
    @classmethod
    def check_si2n(cls, v):
        if not (isinstance(v, float) and v > 0):
            raise ValueError("si2n must be a positive float")
        return v

    @field_validator("p2n")
    @classmethod
    def check_p2n(cls, v):
        if not (isinstance(v, float) and v > 0):
            raise ValueError("p2n must be a positive float")
        return v

    @field_validator("o2no")
    @classmethod
    def check_o2no(cls, v):
        if not (isinstance(v, float) and v > 0):
            raise ValueError("o2no must be a positive float")
        return v

    @field_validator("o2nh")
    @classmethod
    def check_o2nh(cls, v):
        if not (isinstance(v, float) and v > 0):
            raise ValueError("o2nh must be a positive float")
        return v

    @field_validator("c2n")
    @classmethod
    def check_c2n(cls, v):
        if not (isinstance(v, float) and v > 0):
            raise ValueError("c2n must be a positive float")
        return v

    @field_validator("gamman")
    @classmethod
    def check_gamman(cls, v):
        if not (isinstance(v, float) and 0 <= v <= 1):
            raise ValueError("gamman must be a float between 0 and 1")
        return v

    @field_validator("pco2a")
    @classmethod
    def check_pco2a(cls, v):
        if not (isinstance(v, float) and v > 0):
            raise ValueError("pco2a must be a positive float")
        return v

    @field_validator("kmdn")
    @classmethod
    def check_kmdn(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and x >= 0 for x in v)
        ):
            raise ValueError("kmdn must be a list of 2 non-negative floats")
        return v

    @field_validator("kmdsi")
    @classmethod
    def check_kmdsi(cls, v):
        if not (
            isinstance(v, list)
            and len(v) == 2
            and all(isinstance(x, float) and x >= 0 for x in v)
        ):
            raise ValueError("kmdsi must be a list of 2 non-negative floats")
        return v

    @model_validator(mode="after")
    def check_rhoz_sum(self):
        if sum(self.rhoz) != 1:
            raise ValueError("Sum of rhoz values must equal 1")
        return self


class Misc(NamelistBaseModel):
    iws: Optional[int] = Field(
        0,
        description="Flag to enable or disable diatom sinking velocity dependence on NO3 concentration",
    )
    no3c: Optional[float] = Field(
        2.0,
        description="Critical NO3 concentration (mmol/m3) for diatom sinking velocity calculation",
    )
    ws1: Optional[float] = Field(
        2.5,
        description="Diatom sinking velocity (m/day) when NO3 concentration is below no3c",
    )
    ws2: Optional[float] = Field(
        2.0,
        description="Diatom sinking velocity (m/day) when NO3 concentration is above no3c",
    )
    iclam: Optional[int] = Field(
        0, description="Flag to enable or disable clam grazing model"
    )
    deltaz: Optional[int] = Field(
        1, description="Vertical grid size (meter) for clam grazing model"
    )
    kcex: Optional[float] = Field(0.002, description="Clam excretion rate (day^-1)")
    nperclam: Optional[float] = Field(
        0.39032, description="Nitrogen content per clam (mmol[N])"
    )
    wclam: Optional[float] = Field(0.00545, description="Clam weight (g)")
    fclam: Optional[int] = Field(
        40, description="Clam filtration rate (L.g[AFDW]^-1.day^-1)"
    )
    nclam0: Optional[int] = Field(2000, description="Initial number of clams")
    fs2: Optional[list] = Field(
        [0.1, 0.1, 0.8],
        description="Partitioning coefficients for S2 from water column into sediment (3 values)",
    )
    rks2: Optional[list] = Field(
        [0.004, 0.0001, 0.0],
        description="Changing rates of remineralization for sediment S2 (3 values, day^-1)",
    )
    mks2: Optional[list] = Field(
        [0.1, 0.01, 0.0],
        description="Maximum remineralization rates for sediment S2 (3 values, day^-1)",
    )
    fdn: Optional[list] = Field(
        [0.15, 0.1, 0.75],
        description="Partitioning coefficients for DN from water column into sediment (3 values)",
    )
    rkdn: Optional[list] = Field(
        [0.004, 0.0001, 0.0],
        description="Changing rates of remineralization for sediment DN (3 values, day^-1)",
    )
    mkdn: Optional[list] = Field(
        [0.1, 0.01, 0.0],
        description="Maximum remineralization rates for sediment DN (3 values, day^-1)",
    )
    fdsi: Optional[list] = Field(
        [0.3, 0.3, 0.4],
        description="Partitioning coefficients for DSi from water column into sediment (3 values)",
    )
    rkdsi: Optional[list] = Field(
        [0.004, 0.0001, 0.0],
        description="Changing rates of remineralization for sediment DSi (3 values, day^-1)",
    )
    mkdsi: Optional[list] = Field(
        [0.1, 0.01, 0.0],
        description="Maximum remineralization rates for sediment DSi (3 values, day^-1)",
    )

    @field_validator("iws")
    @classmethod
    def validate_iws(cls, v: int) -> int:
        if v not in [0, 1]:
            raise ValueError("iws must be either 0 or 1")
        return v

    @field_validator("no3c")
    @classmethod
    def validate_no3c(cls, v: float) -> float:
        if v <= 0:
            raise ValueError("no3c must be positive")
        return v

    @field_validator("ws1")
    @classmethod
    def validate_ws1(cls, v: float) -> float:
        if v < 0:
            raise ValueError("ws1 must be non-negative")
        return v

    @field_validator("ws2")
    @classmethod
    def validate_ws2(cls, v: float) -> float:
        if v < 0:
            raise ValueError("ws2 must be non-negative")
        return v

    @field_validator("iclam")
    @classmethod
    def validate_iclam(cls, v: int) -> int:
        if v not in [0, 1]:
            raise ValueError("iclam must be either 0 or 1")
        return v

    @field_validator("deltaz")
    @classmethod
    def validate_deltaz(cls, v: float) -> float:
        if v <= 0:
            raise ValueError("deltaz must be positive")
        return v

    @field_validator("kcex")
    @classmethod
    def validate_kcex(cls, v: float) -> float:
        if v < 0 or v > 1:
            raise ValueError("kcex must be between 0 and 1")
        return v

    @field_validator("nperclam")
    @classmethod
    def validate_nperclam(cls, v: float) -> float:
        if v <= 0:
            raise ValueError("nperclam must be positive")
        return v

    @field_validator("wclam")
    @classmethod
    def validate_wclam(cls, v: float) -> float:
        if v <= 0:
            raise ValueError("wclam must be positive")
        return v

    @field_validator("fclam")
    @classmethod
    def validate_fclam(cls, v: float) -> float:
        if v <= 0:
            raise ValueError("fclam must be positive")
        return v

    @field_validator("nclam0")
    @classmethod
    def validate_nclam0(cls, v: int) -> int:
        if v < 0:
            raise ValueError("nclam0 must be non-negative")
        return v

    @field_validator("fs2")
    @classmethod
    def validate_fs2(cls, v: List[float]) -> List[float]:
        if len(v) != 3 or not all(0 <= x <= 1 for x in v) or abs(sum(v) - 1) > 1e-6:
            raise ValueError(
                "fs2 must be a list of 3 values between 0 and 1, summing to 1"
            )
        return v

    @field_validator("rks2")
    @classmethod
    def validate_rks2(cls, v: List[float]) -> List[float]:
        if len(v) != 3 or not all(x >= 0 for x in v):
            raise ValueError("rks2 must be a list of 3 non-negative values")
        return v

    @field_validator("mks2")
    @classmethod
    def validate_mks2(cls, v: List[float]) -> List[float]:
        if len(v) != 3 or not all(x >= 0 for x in v):
            raise ValueError("mks2 must be a list of 3 non-negative values")
        return v

    @field_validator("fdn")
    @classmethod
    def validate_fdn(cls, v: List[float]) -> List[float]:
        if len(v) != 3 or not all(0 <= x <= 1 for x in v) or abs(sum(v) - 1) > 1e-6:
            raise ValueError(
                "fdn must be a list of 3 values between 0 and 1, summing to 1"
            )
        return v

    @field_validator("rkdn")
    @classmethod
    def validate_rkdn(cls, v: List[float]) -> List[float]:
        if len(v) != 3 or not all(x >= 0 for x in v):
            raise ValueError("rkdn must be a list of 3 non-negative values")
        return v

    @field_validator("mkdn")
    @classmethod
    def validate_mkdn(cls, v: List[float]) -> List[float]:
        if len(v) != 3 or not all(x >= 0 for x in v):
            raise ValueError("mkdn must be a list of 3 non-negative values")
        return v

    @field_validator("fdsi")
    @classmethod
    def validate_fdsi(cls, v: List[float]) -> List[float]:
        if len(v) != 3 or not all(0 <= x <= 1 for x in v) or abs(sum(v) - 1) > 1e-6:
            raise ValueError(
                "fdsi must be a list of 3 values between 0 and 1, summing to 1"
            )
        return v

    @field_validator("rkdsi")
    @classmethod
    def validate_rkdsi(cls, v: List[float]) -> List[float]:
        if len(v) != 3 or not all(x >= 0 for x in v):
            raise ValueError("rkdsi must be a list of 3 non-negative values")
        return v

    @field_validator("mkdsi")
    @classmethod
    def validate_mkdsi(cls, v: List[float]) -> List[float]:
        if len(v) != 3 or not all(x >= 0 for x in v):
            raise ValueError("mkdsi must be a list of 3 non-negative values")
        return v


class Cosine(NamelistBaseModel):
    marco: Optional[Marco] = Field(default_factory=Marco)
    core: Optional[Core] = Field(default_factory=Core)
    misc: Optional[Misc] = Field(default_factory=Misc)



# This file was auto generated from a SCHISM namelist file on 2025-01-29.

import re
from datetime import datetime
from pathlib import Path
from typing import Optional

from pydantic import Field, field_validator, model_validator

from rompy_schism.namelists.basemodel import NamelistBaseModel


class Proc(NamelistBaseModel):
    procname: Optional[str] = Field(
        "ACS", description="Project name for the simulation"
    )
    dimmode: Optional[int] = Field(
        2, description="Mode of run: 2 for 2D (always 2D when coupled to SCHISM)"
    )
    lstea: Optional[bool] = Field(
        False, description="Steady mode flag (under development)"
    )
    lqstea: Optional[bool] = Field(
        False,
        description="Quasi-Steady Mode flag. If True, WWM-II performs subiterations defined as DELTC/NQSITER unless QSCONVI is reached",
    )
    lsphe: Optional[bool] = Field(
        True, description="Flag for spherical coordinates (lon/lat)"
    )
    lnautin: Optional[bool] = Field(
        True,
        description="Flag for nautical convention in input angles (degrees). Recommended to be True",
    )
    lnautout: Optional[bool] = Field(
        True,
        description="Flag for nautical convention in output angles. If True, 0 is from north, 90 is from east. If False, mathematical convention is used (0: to east, 90: to north)",
    )
    lmono_in: Optional[bool] = Field(
        False,
        description="Flag for prescribing monochromatic wave height Hmono as boundary conditions. Incident wave is defined as Hmono = sqrt(2) * Hs",
    )
    lmono_out: Optional[bool] = Field(
        False, description="Flag for outputting wave heights in terms of Lmono"
    )
    begtc: Optional[str] = Field(
        "20200101.000000",
        description="Start time of the simulation in format 'yyyymmdd.hhmmss'",
    )
    deltc: Optional[int] = Field(
        600, description="Time step in seconds (must match dt*nstep_wwm in SCHISM)"
    )
    unitc: Optional[str] = Field(
        "SEC", description="Unit of time step (SEC for seconds)"
    )
    endtc: Optional[str] = Field(
        "20200201.000000",
        description="End time of the simulation in format 'yyyymmdd.hhmmss'",
    )
    dmin: Optional[float] = Field(
        0.01, description="Minimum water depth. Must be the same as h0 in SCHISM"
    )

    @field_validator("procname")
    def validate_procname(cls, v):
        if not v.strip():
            raise ValueError("Project name cannot be empty")
        return v

    @field_validator("dimmode")
    def validate_dimmode(cls, v):
        if v != 2:
            raise ValueError("DIMMODE must be 2 when coupled to SCHISM")
        return v

    @field_validator("lstea")
    def validate_lstea(cls, v):
        return v

    @field_validator("lqstea")
    def validate_lqstea(cls, v):
        return v

    @field_validator("lsphe")
    def validate_lsphe(cls, v):
        return v

    @field_validator("lnautin")
    def validate_lnautin(cls, v):
        return v

    @field_validator("lnautout")
    def validate_lnautout(cls, v):
        return v

    @field_validator("lmono_in")
    def validate_lmono_in(cls, v):
        return v

    @field_validator("lmono_out")
    def validate_lmono_out(cls, v):
        return v

    @field_validator("begtc")
    def validate_begtc(cls, v):
        try:
            datetime.strptime(v, "%Y%m%d.%H%M%S")
        except ValueError:
            raise ValueError("Invalid date format for BEGTC")
        return v

    @field_validator("deltc")
    def validate_deltc(cls, v):
        if v <= 0:
            raise ValueError("DELTC must be positive")
        return v

    @field_validator("unitc")
    def validate_unitc(cls, v):
        if v.upper() != "SEC":
            raise ValueError("UNITC must be SEC")
        return v.upper()

    @field_validator("endtc")
    def validate_endtc(cls, v):
        try:
            datetime.strptime(v, "%Y%m%d.%H%M%S")
        except ValueError:
            raise ValueError("Invalid date format for ENDTC")
        return v

    @field_validator("dmin")
    def validate_dmin(cls, v):
        if v <= 0:
            raise ValueError("DMIN must be positive")
        return v


class Coupl(NamelistBaseModel):
    lcpl: Optional[bool] = Field(
        True,
        description="Main switch to enable coupling with the current model. Should be kept on for SCHISM-WWM coupling.",
    )
    radflag: Optional[str] = Field(
        "LON",
        description="Determines the formulation for wave-induced forces. 'LON' for Longuet-Higgins formulation, 'VOR' for vortex formulation. Usually set to 'LON'.",
    )
    letot: Optional[bool] = Field(
        False,
        description="Option to compute wave-induced radiation stress. If True, radiation stress is based on the integrated wave spectrum. If False (recommended), it's estimated using the directional spectra as described in Roland et al. (2008). False is preferred as it preserves spectral information.",
    )
    nlvt: Optional[int] = Field(
        10, description="Number of vertical layers. Not used when coupled with SCHISM."
    )
    dtcoup: Optional[float] = Field(
        600.0,
        description="Coupling time step in seconds. Not used when coupled to SCHISM.",
    )

    @field_validator("lcpl")
    @classmethod
    def validate_lcpl(cls, v):
        return v

    @field_validator("radflag")
    @classmethod
    def validate_radflag(cls, v):
        if v not in ["LON", "VOR"]:
            raise ValueError("radflag must be either 'LON' or 'VOR'")
        return v

    @field_validator("letot")
    @classmethod
    def validate_letot(cls, v):
        return v

    @field_validator("nlvt")
    @classmethod
    def validate_nlvt(cls, v):
        if v <= 0:
            raise ValueError("nlvt must be a positive integer")
        return v

    @field_validator("dtcoup")
    @classmethod
    def validate_dtcoup(cls, v):
        if v <= 0:
            raise ValueError("dtcoup must be a positive number")
        return v


class Grid(NamelistBaseModel):
    lcird: Optional[bool] = Field(
        True,
        description="Flag to indicate if a full circle in directional space is used. If True, MINDIR and MAXDIR are ignored.",
    )
    lstag: Optional[bool] = Field(
        False,
        description="Flag to stagger directional bins with a half Dtheta. Can only be True for regular grids to avoid characteristic line aligning with grid line.",
    )
    mindir: Optional[float] = Field(
        0.0,
        description="Minimum direction for simulation in degrees (nautical convention; 0: from N; 90: from E). Not used if LCIRD is True.",
    )
    maxdir: Optional[float] = Field(
        360.0,
        description="Maximum direction for simulation in degrees. May be less than MINDIR. Not used if LCIRD is True.",
    )
    mdc: Optional[int] = Field(24, description="Number of directional bins")
    frlow: Optional[float] = Field(
        0.04,
        description="Low frequency limit of the discrete wave period in Hz (1/period)",
    )
    frhigh: Optional[float] = Field(
        1.0, description="High frequency limit of the discrete wave period in Hz"
    )
    msc: Optional[int] = Field(24, description="Number of frequency bins")
    filegrid: Optional[str] = Field(
        "hgrid_WWM.gr3",
        description="Name of the grid file. Should be 'hgridi_WWM.gr3' if IGRIDTYPE is 3 (SCHISM)",
    )
    igridtype: Optional[int] = Field(
        3,
        description="Grid type used. 1: XFN system.dat, 2: WWM-PERIODIC, 3: SCHISM, 4: old WWM type",
    )
    lslop: Optional[bool] = Field(
        False, description="Flag to enable bottom slope limiter"
    )
    slmax: Optional[float] = Field(
        0.2, description="Maximum allowed bottom slope when LSLOP is True"
    )
    lvar1d: Optional[bool] = Field(
        False,
        description="Flag to use variable dx in 1D mode. Not used with SCHISM (IGRIDTYPE = 3)",
    )
    loptsig: Optional[bool] = Field(
        False,
        description="Flag to use optimal distributions of frequencies in spectral space (fi+1 = fi * 1.1)",
    )

    @field_validator("lcird")
    def validate_lcird(cls, v):
        return v

    @field_validator("lstag")
    def validate_lstag(cls, v, info):
        if v and info.data.get("igridtype") != 1:
            raise ValueError("LSTAG can only be True for regular grids (IGRIDTYPE = 1)")
        return v

    @field_validator("mindir")
    def validate_mindir(cls, v):
        if not 0 <= v < 360:
            raise ValueError("MINDIR must be between 0 and 360 degrees")
        return v

    @field_validator("maxdir")
    def validate_maxdir(cls, v):
        if not 0 <= v <= 360:
            raise ValueError("MAXDIR must be between 0 and 360 degrees")
        return v

    @field_validator("mdc")
    def validate_mdc(cls, v):
        if v <= 0:
            raise ValueError("MDC must be a positive integer")
        return v

    @field_validator("frlow")
    def validate_frlow(cls, v):
        if v <= 0:
            raise ValueError("FRLOW must be a positive number")
        return v

    @field_validator("frhigh")
    def validate_frhigh(cls, v, info):
        if v <= info.data.get("frlow", 0):
            raise ValueError("FRHIGH must be greater than FRLOW")
        return v

    @field_validator("msc")
    def validate_msc(cls, v):
        if v <= 0:
            raise ValueError("MSC must be a positive integer")
        return v

    @field_validator("filegrid")
    def validate_filegrid(cls, v):
        if not v.strip():
            raise ValueError("FILEGRID must not be empty")
        return v

    @field_validator("igridtype")
    def validate_igridtype(cls, v):
        if v not in [1, 2, 3, 4]:
            raise ValueError("IGRIDTYPE must be 1, 2, 3, or 4")
        return v

    @field_validator("lslop")
    def validate_lslop(cls, v):
        return v

    @field_validator("slmax")
    def validate_slmax(cls, v):
        if v <= 0:
            raise ValueError("SLMAX must be a positive number")
        return v

    @field_validator("lvar1d")
    def validate_lvar1d(cls, v, info):
        if v and info.data.get("igridtype") == 3:
            raise ValueError("LVAR1D cannot be True when IGRIDTYPE is 3 (SCHISM)")
        return v

    @field_validator("loptsig")
    def validate_loptsig(cls, v):
        return v


class Init(NamelistBaseModel):
    lhotr: Optional[bool] = Field(
        False,
        description="Flag to indicate whether to use a hotstart file for initial conditions. If True, the model will read initial conditions from a file specified in the HOTFILE section.",
    )
    linid: Optional[bool] = Field(
        True,
        description="Flag to control the initial condition setup. If False, default initial conditions are used. If True, it allows for using external sources (e.g., WW3) as initial conditions.",
    )
    initstyle: Optional[int] = Field(
        2,
        description="Specifies the method for initializing wave conditions. 1 for Parametric Jonswap, 2 for reading from Global NETCDF files. Option 2 only works if IBOUNDFORMAT is set to 3.",
    )

    # validator to ensure thatn only one of lhotr and linid is True
    @model_validator(mode="after")
    def check_init(self):
        if self.lhotr and self.linid:
            raise ValueError("Only one of LHOTR and LINID can be True")
        return self

    @field_validator("initstyle")
    @classmethod
    def validate_initstyle(cls, v):
        if v not in [1, 2]:
            raise ValueError("initstyle must be either 1 or 2")
        return v


class Bouc(NamelistBaseModel):
    lbcse: Optional[bool] = Field(
        True, description="Determines if wave boundary data is time-dependent."
    )
    lbinter: Optional[bool] = Field(
        True,
        description="Controls time interpolation when LBCSE is true. Not available for quasi-steady mode within subtime steps.",
    )
    lbcwa: Optional[bool] = Field(False, description="Enables parametric wave spectra.")
    lbcsp: Optional[bool] = Field(
        True,
        description="Enables specification of non-parametric wave spectra, defined in FILEWAVE.",
    )
    linhom: Optional[bool] = Field(
        True, description="Enables non-uniform wave boundary conditions in space."
    )
    lbsp1d: Optional[bool] = Field(
        False,
        description="Specifies 1D (frequency space only) format for FILEWAVE when LBCSP is true and LINHOM is false.",
    )
    lbsp2d: Optional[bool] = Field(
        True,
        description="Specifies 2D format for FILEWAVE when LBCSP is true and LINHOM is false.",
    )
    begtc: Optional[str] = Field(
        "20200101.000000",
        description="Start time of the wave boundary file (FILEWAVE) in 'YYYYMMDD.HHMMSS' format.",
    )
    deltc: Optional[int] = Field(1, description="Time step in FILEWAVE.")
    unitc: Optional[str] = Field(
        "HR", description="Time unit for DELTC. Can be HR, MIN, or SEC."
    )
    endtc: Optional[str] = Field(
        "20200201.000000",
        description="End time of the wave boundary file in 'YYYYMMDD.HHMMSS' format.",
    )
    filebound: Optional[str] = Field(
        "wwmbnd.gr3",
        description="Boundary file defining boundary conditions and Neumann nodes.",
    )
    iboundformat: Optional[int] = Field(
        6,
        description="Format of the boundary file. 1: WWM, 3: WW3 (2D spectra in netcdf format only - LBCWA=T), 6: WW3 2D spectra in netcdf format with LBCSP=T.",
    )
    filewave: Optional[str | Path] = Field(
        "ww3.acs.SCHISM.nc",
        description="Boundary file defining boundary input from WW3. This will be generated by rompy if wave data is specified in the data input",
    )
    lindsprdeg: Optional[bool] = Field(
        True,
        description="For 1D wave spectra, defines whether directional spreading input is in degrees (true) or exponent (false).",
    )
    lparmdir: Optional[bool] = Field(
        False,
        description="If true, directional spreading is read from WBDS in exponential format. Only valid for 1D spectra.",
    )
    wbhs: Optional[float] = Field(
        2.0,
        description="Significant wave height at the boundary for parametric spectra.",
    )
    wbss: Optional[int] = Field(
        2,
        description="Spectral shape parameter. 1/-1: Pierson-Moskowitz, 2/-2: JONSWAP, 3/-3: all in one BIN, 4: Gauss. Sign determines if WBTP is peak (+) or mean period (-).",
    )
    wbtp: Optional[float] = Field(
        8.0,
        description="Peak or mean wave period at the boundary (seconds), depending on the sign of WBSS.",
    )
    wbdm: Optional[float] = Field(
        90.0, description="Average wave direction at the boundary (degrees)."
    )
    wbdsms: Optional[int] = Field(
        1, description="Directional spreading value format. 1: degrees, 2: exponent."
    )
    wbds: Optional[float] = Field(
        10.0, description="Directional spreading at the boundary (degrees or exponent)."
    )
    wbgauss: Optional[float] = Field(
        0.1, description="Factor for Gaussian distribution if WBSS=4."
    )
    wbpken: Optional[float] = Field(
        3.3, description="Peak enhancement factor for JONSWAP spectra if WBSS=2."
    )

    @field_validator("lbcse")
    @classmethod
    def validate_lbcse(cls, v):
        return v

    @field_validator("lbinter")
    @classmethod
    def validate_lbinter(cls, v, info):
        if v and not info.data.get("lbcse"):
            raise ValueError("LBINTER can only be True if LBCSE is True")
        return v

    @field_validator("lbcwa")
    @classmethod
    def validate_lbcwa(cls, v):
        return v

    @field_validator("lbcsp")
    @classmethod
    def validate_lbcsp(cls, v):
        return v

    @field_validator("linhom")
    @classmethod
    def validate_linhom(cls, v):
        return v

    # Not clear from sample inputs wheither these are strict requirements, or
    # are ignored if not met. Leaving out for now
    # @field_validator("lbsp1d")
    # @classmethod
    # def validate_lbsp1d(cls, v, info):
    #     if v and (not info.data.get("lbcsp") or info.data.get("linhom")):
    #         raise ValueError(
    #             "LBSP1D can only be True if LBCSP is True and LINHOM is False"
    #         )
    #     return v
    #
    # @field_validator("lbsp2d")
    # @classmethod
    # def validate_lbsp2d(cls, v, info):
    #     if v and (not info.data.get("lbcsp") or info.data.get("linhom")):
    #         raise ValueError(
    #             "LBSP2D can only be True if LBCSP is True and LINHOM is False"
    #         )
    #     return v

    @field_validator("begtc")
    @classmethod
    def validate_begtc(cls, v):
        if not re.match(r"\d{8}\.\d{6}", v):
            raise ValueError("BEGTC must be in YYYYMMDD.HHMMSS format")
        return v

    @field_validator("deltc")
    @classmethod
    def validate_deltc(cls, v):
        if v <= 0:
            raise ValueError("DELTC must be positive")
        return v

    @field_validator("unitc")
    @classmethod
    def validate_unitc(cls, v):
        if v not in ["HR", "MIN", "SEC"]:
            raise ValueError("UNITC must be HR, MIN, or SEC")
        return v

    @field_validator("endtc")
    @classmethod
    def validate_endtc(cls, v):
        if not re.match(r"\d{8}\.\d{6}", v):
            raise ValueError("ENDTC must be in YYYYMMDD.HHMMSS format")
        return v

    @field_validator("filebound")
    @classmethod
    def validate_filebound(cls, v):
        if not v.endswith(".gr3"):
            raise ValueError("FILEBOUND must have a .gr3 extension")
        return v

    @field_validator("iboundformat")
    @classmethod
    def validate_iboundformat(cls, v):
        if v not in [1, 3, 6]:
            raise ValueError("IBOUNDFORMAT must be 1, 3, or 6")
        return v

    @field_validator("filewave")
    @classmethod
    def validate_filewave(cls, v):
        if not str(v).endswith(".nc"):
            raise ValueError("FILEWAVE must have a .nc extension")
        return v

    @field_validator("lindsprdeg")
    @classmethod
    def validate_lindsprdeg(cls, v):
        return v

    @field_validator("lparmdir")
    @classmethod
    def validate_lparmdir(cls, v):
        return v

    @field_validator("wbhs")
    @classmethod
    def validate_wbhs(cls, v):
        if v <= 0:
            raise ValueError("WBHS must be positive")
        return v

    @field_validator("wbss")
    @classmethod
    def validate_wbss(cls, v):
        if v not in [-3, -2, -1, 1, 2, 3, 4]:
            raise ValueError("WBSS must be -3, -2, -1, 1, 2, 3, or 4")
        return v

    @field_validator("wbtp")
    @classmethod
    def validate_wbtp(cls, v):
        if v <= 0:
            raise ValueError("WBTP must be positive")
        return v

    @field_validator("wbdm")
    @classmethod
    def validate_wbdm(cls, v):
        if not 0 <= v <= 360:
            raise ValueError("WBDM must be between 0 and 360")
        return v

    @field_validator("wbdsms")
    @classmethod
    def validate_wbdsms(cls, v):
        if v not in [1, 2]:
            raise ValueError("WBDSMS must be 1 or 2")
        return v

    @field_validator("wbds")
    @classmethod
    def validate_wbds(cls, v):
        if v < 0:
            raise ValueError("WBDS must be non-negative")
        return v

    @field_validator("wbgauss")
    @classmethod
    def validate_wbgauss(cls, v, info):
        if info.data.get("wbss") == 4 and v <= 0:
            raise ValueError("WBGAUSS must be positive when WBSS is 4")
        return v

    @field_validator("wbpken")
    @classmethod
    def validate_wbpken(cls, v, info):
        if abs(info.data.get("wbss", 0)) == 2 and v <= 1:
            raise ValueError("WBPKEN must be greater than 1 when WBSS is 2 or -2")
        return v


class Engs(NamelistBaseModel):
    mesnl: Optional[int] = Field(
        1,
        description="Controls the nonlinear wave-wave interactions (NL4) using the Discrete Interaction Approximation. 1 enables the interactions, 0 disables them.",
    )
    mesin: Optional[int] = Field(
        1,
        description="Specifies the wind input formulation. Options include: 0 (no wind), 1 (Ardhuin et al.), 2 (ECMWF physics), 3 (Makin Stam), 4 (Babanin et al.), 5 (Cycle 3).",
    )
    ifric: Optional[int] = Field(
        1,
        description="Determines the formulation for the atmospheric boundary layer. Should be 1 when MESIN=1, and 4 when MESIN=3.",
    )
    mesbf: Optional[int] = Field(
        2,
        description="Selects the bottom friction formulation. 1 for JONSWAP (Default), 2 for Madsen et al. (1989), 3 for SHOWEX.",
    )
    fricc: Optional[float] = Field(
        0.006,
        description="Bottom friction coefficient or roughness, depending on MESBF. For MESBF=1: JONSWAP coefficient [0.038,0.067]. For MESBF=2: physical bottom roughness. For MESBF=3: D50 (negative value reads from SHOWEX_D50.gr3).",
    )
    mesbr: Optional[int] = Field(
        1, description="Enables (1) or disables (0) shallow water wave breaking."
    )
    ibreak: Optional[int] = Field(
        1,
        description="Selects the wave breaking formulation. Options range from 1 to 6, each representing a different model or approach.",
    )
    icrit: Optional[int] = Field(
        1,
        description="Specifies the wave breaking criterion. Options 1-6 represent different methods for determining the breaking point.",
    )
    brcr: Optional[float] = Field(
        0.73,
        description="Breaking criterion parameter. Its meaning depends on IBREAK and ICRIT settings.",
    )
    a_brcr: Optional[float] = Field(
        0.76,
        description="Coefficient used in ICRIT=4,5 for calculating the breaking criterion.",
    )
    b_brcr: Optional[float] = Field(
        0.29,
        description="Coefficient used in ICRIT=4,5 for calculating the breaking criterion.",
    )
    min_brcr: Optional[float] = Field(
        0.25, description="Minimum value for the breaking criterion when ICRIT=4,5."
    )
    max_brcr: Optional[float] = Field(
        0.8, description="Maximum value for the breaking criterion when ICRIT=4,5."
    )
    a_biph: Optional[float] = Field(
        0.2, description="Biphase coefficient, used when IBREAK=3."
    )
    br_coef_method: Optional[int] = Field(
        1,
        description="Method for determining the breaking coefficient. 1 for constant, 2 for adaptive.",
    )
    b_alp: Optional[float] = Field(
        0.5,
        description="Breaking coefficient. If BR_COEF_METHOD = 2, B_ALP should be around 40.",
    )
    zprof_break: Optional[int] = Field(
        2,
        description="Specifies the vertical distribution function of the wave breaking source term in 3D runs. Options 1-6 represent different distribution functions.",
    )
    bc_break: Optional[int] = Field(
        1,
        description="Controls the application of depth-limited breaking at boundaries. 1 to enable, 0 to disable.",
    )
    iroller: Optional[int] = Field(
        0,
        description="Enables (1) or disables (0) the wave roller model. Currently not in use.",
    )
    alprol: Optional[float] = Field(
        0.85,
        description="Alpha coefficient for the wave roller model, determining the energy transfer to the roller. Range: 0 to 1.",
    )
    meveg: Optional[int] = Field(
        0,
        description="Enables (1) or disables (0) vegetation effects. If enabled, isav must be 1 in param.nml.",
    )
    lmaxetot: Optional[bool] = Field(
        True,
        description="Controls the use of wave breaking limiter to limit shallow water wave height. True to enable, False to disable.",
    )
    mesds: Optional[int] = Field(
        1,
        description="Specifies the formulation for the whitecapping source function. Should have the same value as MESIN.",
    )
    mestr: Optional[int] = Field(
        1,
        description="Selects the formulation for triad 3 wave interactions. 0 (off), 1 (Lumped Triad Approx.), 2 (corrected LTA by Salmon et al. (2016)).",
    )
    trico: Optional[float] = Field(
        0.1,
        description="Proportionality constant (α_EB) for triad interactions. Default is 0.1.",
    )
    trira: Optional[float] = Field(
        2.5,
        description="Ratio of maximum frequency considered in triads over mean frequency. Suggested value is 2.5.",
    )
    triurs: Optional[float] = Field(
        0.1,
        description="Critical Ursell number for triad computations. Triads are not computed if Ursell number < TRIURS.",
    )

    @field_validator("mesnl")
    @classmethod
    def validate_mesnl(cls, v):
        if v not in [0, 1]:
            raise ValueError("MESNL must be either 0 or 1")
        return v

    @field_validator("mesin")
    @classmethod
    def validate_mesin(cls, v):
        if v not in range(6):
            raise ValueError("MESIN must be between 0 and 5")
        return v

    @field_validator("ifric")
    @classmethod
    def validate_ifric(cls, v, info):
        if info.data.get("mesin") == 1 and v != 1:
            raise ValueError("IFRIC should be 1 when MESIN is 1")
        elif info.data.get("mesin") == 3 and v != 4:
            raise ValueError("IFRIC should be 4 when MESIN is 3")
        return v

    @field_validator("mesbf")
    @classmethod
    def validate_mesbf(cls, v):
        if v not in [1, 2, 3]:
            raise ValueError("MESBF must be 1, 2, or 3")
        return v

    @field_validator("fricc")
    @classmethod
    def validate_fricc(cls, v, info):
        if info.data.get("mesbf") == 1 and not 0.038 <= v <= 0.067:
            raise ValueError("FRICC must be between 0.038 and 0.067 when MESBF is 1")
        return v

    @field_validator("mesbr")
    @classmethod
    def validate_mesbr(cls, v):
        if v not in [0, 1]:
            raise ValueError("MESBR must be either 0 or 1")
        return v

    @field_validator("ibreak")
    @classmethod
    def validate_ibreak(cls, v):
        if v not in range(1, 7):
            raise ValueError("IBREAK must be between 1 and 6")
        return v

    @field_validator("icrit")
    @classmethod
    def validate_icrit(cls, v):
        if v not in range(1, 7):
            raise ValueError("ICRIT must be between 1 and 6")
        return v

    @field_validator("brcr")
    @classmethod
    def validate_brcr(cls, v, info):
        if info.data.get("ibreak") in [1, 5] and v != 0.73:
            print("Warning: Default BRCR for IBREAK 1 or 5 is 0.73")
        elif info.data.get("ibreak") in [2, 3] and v != 0.42:
            print("Warning: Default BRCR for IBREAK 2 or 3 is 0.42")
        elif info.data.get("ibreak") == 4 and v != -1.3963:
            print("Warning: Default BRCR for IBREAK 4 is -1.3963")
        return v

    @field_validator("br_coef_method")
    @classmethod
    def validate_br_coef_method(cls, v):
        if v not in [1, 2]:
            raise ValueError("BR_COEF_METHOD must be either 1 or 2")
        return v

    @field_validator("b_alp")
    @classmethod
    def validate_b_alp(cls, v, info):
        if info.data.get("br_coef_method") == 2 and v != 40:
            print("Warning: B_ALP should be around 40 when BR_COEF_METHOD is 2")
        return v

    @field_validator("zprof_break")
    @classmethod
    def validate_zprof_break(cls, v):
        if v not in range(1, 7):
            raise ValueError("ZPROF_BREAK must be between 1 and 6")
        return v

    @field_validator("bc_break")
    @classmethod
    def validate_bc_break(cls, v):
        if v not in [0, 1]:
            raise ValueError("BC_BREAK must be either 0 or 1")
        return v

    @field_validator("iroller")
    @classmethod
    def validate_iroller(cls, v):
        if v not in [0, 1]:
            raise ValueError("IROLLER must be either 0 or 1")
        return v

    @field_validator("alprol")
    @classmethod
    def validate_alprol(cls, v):
        if not 0 <= v <= 1:
            raise ValueError("ALPROL must be between 0 and 1")
        return v

    @field_validator("meveg")
    @classmethod
    def validate_meveg(cls, v):
        if v not in [0, 1]:
            raise ValueError("MEVEG must be either 0 or 1")
        return v

    @field_validator("mesds")
    @classmethod
    def validate_mesds(cls, v, info):
        if v != info.data.get("mesin"):
            raise ValueError("MESDS should have the same value as MESIN")
        return v

    @field_validator("mestr")
    @classmethod
    def validate_mestr(cls, v):
        if v not in [0, 1, 2]:
            raise ValueError("MESTR must be 0, 1, or 2")
        return v


class Nums(NamelistBaseModel):
    icomp: Optional[int] = Field(
        3,
        description="Controls the integration scheme for splitting and advection. 0: All explicit. 1: Implicit geographical advection. 2: Implicit advection and semi-implicit source terms. 3: Fully implicit, no splitting.",
    )
    amethod: Optional[int] = Field(
        7,
        description="Controls the advection method in geographical space. Values 0-7 represent different schemes, including explicit, implicit, and PETSc-based methods.",
    )
    smethod: Optional[int] = Field(
        6,
        description="Controls the integration method for source terms. 0: No source terms. 1-6: Various splitting and integration schemes.",
    )
    dmethod: Optional[int] = Field(
        2,
        description="Controls the numerical method in directional space. 0: No advection. 1-4: Various schemes including Crank-Nicholson, Ultimate Quickest, RK5-WENO, and Explicit FVM Upwind.",
    )
    rtheta: Optional[float] = Field(
        0.5,
        description="Weighing factor for DMETHOD = 1. Only useful for Crank Nicholson integration with CFL <= 2.",
    )
    litersplit: Optional[bool] = Field(
        False,
        description="Splitting method. True: double Strang split. False: simple split (more efficient).",
    )
    lfilterth: Optional[bool] = Field(
        False,
        description="Use a CFL filter to limit the advection velocity in directional space. Similar to WW3, but mostly unused as WWMII is always stable.",
    )
    maxcflth: Optional[float] = Field(
        1.0,
        description="Maximum CFL number in Theta space. Used only if LFILTERTH=True.",
    )
    fmethod: Optional[int] = Field(
        1,
        description="Controls the numerical method in frequency space. 0: No advection. 1: Ultimate Quickest as in WW3 (best).",
    )
    lfiltersig: Optional[bool] = Field(
        False,
        description="Limit the advection velocity in frequency space. Usually False.",
    )
    maxcflsig: Optional[float] = Field(
        1.0,
        description="Maximum CFL number in frequency space. Used only if LFILTERSIG=True.",
    )
    llimt: Optional[bool] = Field(
        True, description="Switch on/off Action limiter. Must mostly be turned on."
    )
    melim: Optional[int] = Field(
        1,
        description="Formulation for the action limiter. 1: WAM group (1988). 2: Hersbach Janssen (1999). 3: For Cycle 4 formulation.",
    )
    limfak: Optional[float] = Field(
        0.1,
        description="Proportionality coefficient for the action limiter. MAX_DAC_DT = Limfak * Limiter.",
    )
    ldifr: Optional[bool] = Field(
        False,
        description="Use phase decoupled diffraction approximation. Usually True; if crash, use False.",
    )
    idiffr: Optional[int] = Field(
        1,
        description="Extended WAE accounting for higher order effects. 1: Holthuijsen et al. 2: Liau et al. 3: Toledo et al.",
    )
    lconv: Optional[bool] = Field(
        False,
        description="Estimate convergence criteria and write to disk (quasi-steady - qstea.out).",
    )
    lcfl: Optional[bool] = Field(
        False, description="Write out CFL numbers. Use False to save time."
    )
    nqsiter: Optional[int] = Field(
        1,
        description="Number of quasi-steady (Q-S) sub-divisions within each WWM time step.",
    )
    qsconv1: Optional[float] = Field(
        0.98,
        description="Fraction of grid points that must fulfill absolute wave height criteria EPSH1.",
    )
    qsconv2: Optional[float] = Field(
        0.98,
        description="Fraction of grid points that must fulfill relative wave height criteria EPSH2.",
    )
    qsconv3: Optional[float] = Field(
        0.98,
        description="Fraction of grid points that must fulfill sum. rel. wave action criteria EPSH3.",
    )
    qsconv4: Optional[float] = Field(
        0.98,
        description="Fraction of grid points that must fulfill rel. avg. wave steepness criteria EPSH4.",
    )
    qsconv5: Optional[float] = Field(
        0.98,
        description="Fraction of grid points that must fulfill avg. rel. wave period criteria EPSH5.",
    )
    lexpimp: Optional[bool] = Field(
        False,
        description="Use implicit schemes for frequencies lower than FREQEXP. Used only if ICOMP=0.",
    )
    freqexp: Optional[float] = Field(
        0.1,
        description="Minimum frequency for explicit schemes. Only used if LEXPIMP=True and ICOMP=0.",
    )
    epsh1: Optional[float] = Field(
        0.01, description="Convergence criteria for relative wave height."
    )
    epsh2: Optional[float] = Field(
        0.01, description="Convergence criteria for absolute wave height."
    )
    epsh3: Optional[float] = Field(
        0.01, description="Convergence criteria for the relative sum of wave action."
    )
    epsh4: Optional[float] = Field(
        0.01,
        description="Convergence criteria for the relative average wave steepness.",
    )
    epsh5: Optional[float] = Field(
        0.01, description="Convergence criteria for the relative average wave period."
    )
    lvector: Optional[bool] = Field(
        False,
        description="Use optimized propagation routines for large high performance computers. Try False first.",
    )
    ivector: Optional[int] = Field(
        2,
        description="Used if LVECTOR=True. Different flavors of communications and propagation styles.",
    )
    ladvtest: Optional[bool] = Field(
        False, description="For testing the advection schemes."
    )
    lchkconv: Optional[bool] = Field(
        False,
        description="Needs to be set to True for quasi-steady mode to compute and check the QSCONVi criteria.",
    )
    dtmin_dyn: Optional[float] = Field(
        1.0,
        description="Minimum time step (seconds) for dynamic integration. Controls the smallest time step for triads in SMETHOD.",
    )
    ndyniter: Optional[int] = Field(
        100,
        description="Maximum iterations for dynamic scheme before limiter is applied in the last step.",
    )
    dtmin_sin: Optional[float] = Field(
        1.0,
        description="Minimum time step for the full fractional step method, where each source term is integrated with its own fractional step.",
    )
    dtmin_snl4: Optional[float] = Field(
        1.0,
        description="Minimum time step for SNL4 source term in fractional step method.",
    )
    dtmin_sds: Optional[float] = Field(
        1.0,
        description="Minimum time step for SDS source term in fractional step method.",
    )
    dtmin_snl3: Optional[float] = Field(
        1.0,
        description="Minimum time step for SNL3 source term in fractional step method.",
    )
    dtmin_sbr: Optional[float] = Field(
        0.1,
        description="Minimum time step for SBR source term in fractional step method.",
    )
    dtmin_sbf: Optional[float] = Field(
        1.0,
        description="Minimum time step for SBF source term in fractional step method.",
    )
    ndyniter_sin: Optional[int] = Field(
        10,
        description="Maximum iterations for SIN source term in fractional step approach.",
    )
    ndyniter_snl4: Optional[int] = Field(
        10,
        description="Maximum iterations for SNL4 source term in fractional step approach.",
    )
    ndyniter_sds: Optional[int] = Field(
        10,
        description="Maximum iterations for SDS source term in fractional step approach.",
    )
    ndyniter_sbr: Optional[int] = Field(
        10,
        description="Maximum iterations for SBR source term in fractional step approach.",
    )
    ndyniter_snl3: Optional[int] = Field(
        10,
        description="Maximum iterations for SNL3 source term in fractional step approach.",
    )
    ndyniter_sbf: Optional[int] = Field(
        10,
        description="Maximum iterations for SBF source term in fractional step approach.",
    )
    lsoubound: Optional[bool] = Field(
        False,
        description="Do source terms on boundary. Useful for harbor studies and flume experiments.",
    )
    wae_solverthr: Optional[float] = Field(
        1e-06,
        description="Threshold for the Block-Jacobi or Block-Gauss-Seider solver.",
    )
    maxiter: Optional[int] = Field(
        1000, description="Maximum number of iterations for solver."
    )
    pmin: Optional[float] = Field(
        1.0, description="Maximum percentage of non-converged grid points allowed."
    )
    lnaninfchk: Optional[bool] = Field(
        False, description="Check for NaN and INF. Usually turned off for efficiency."
    )
    lzeta_setup: Optional[bool] = Field(
        False, description="Compute wave setup (simple momentum equation)."
    )
    zeta_meth: Optional[int] = Field(
        0, description="Method for wave setup calculation."
    )
    lsourceswam: Optional[bool] = Field(
        False, description="Use ECMWF WAM formulation for deep water physics."
    )
    block_gauss_seidel: Optional[bool] = Field(
        True,
        description="Use Gauss-Seidel method on each computer block. Faster and uses less memory, but iterations depend on number of processors.",
    )
    lnonl: Optional[bool] = Field(
        False,
        description="Solve the nonlinear system using simpler algorithm (Patankar).",
    )
    aspar_local_level: Optional[int] = Field(
        0,
        description="ASPAR locality level. Controls memory allocation and optimization strategies.",
    )
    l_solver_norm: Optional[bool] = Field(
        False,
        description="Compute solver norm ||A*x-b|| as termination check of Jacobi-Gauss-Seidel solver. Increases cost if True.",
    )
    laccel: Optional[bool] = Field(False, description="Enable acceleration for solver.")

    @field_validator("icomp")
    @classmethod
    def check_icomp(cls, v):
        if v not in [0, 1, 2, 3]:
            raise ValueError("ICOMP must be 0, 1, 2, or 3")
        return v

    @field_validator("amethod")
    @classmethod
    def check_amethod(cls, v):
        if v not in range(8):
            raise ValueError("AMETHOD must be between 0 and 7")
        return v

    @field_validator("smethod")
    @classmethod
    def check_smethod(cls, v):
        if v not in range(7):
            raise ValueError("SMETHOD must be between 0 and 6")
        return v

    @field_validator("dmethod")
    @classmethod
    def check_dmethod(cls, v):
        if v not in range(5):
            raise ValueError("DMETHOD must be between 0 and 4")
        return v

    @field_validator("rtheta")
    @classmethod
    def check_rtheta(cls, v):
        if not 0 <= v <= 1:
            raise ValueError("RTHETA must be between 0 and 1")
        return v

    @field_validator("maxcflth")
    @classmethod
    def check_maxcflth(cls, v):
        if v <= 0:
            raise ValueError("MAXCFLTH must be positive")
        return v

    @field_validator("fmethod")
    @classmethod
    def check_fmethod(cls, v):
        if v not in [0, 1]:
            raise ValueError("FMETHOD must be 0 or 1")
        return v

    @field_validator("maxcflsig")
    @classmethod
    def check_maxcflsig(cls, v):
        if v <= 0:
            raise ValueError("MAXCFLSIG must be positive")
        return v

    @field_validator("melim")
    @classmethod
    def check_melim(cls, v):
        if v not in [1, 2, 3]:
            raise ValueError("MELIM must be 1, 2, or 3")
        return v

    @field_validator("limfak")
    @classmethod
    def check_limfak(cls, v):
        if v <= 0:
            raise ValueError("LIMFAK must be positive")
        return v

    @field_validator("idiffr")
    @classmethod
    def check_idiffr(cls, v):
        if v not in [1, 2, 3]:
            raise ValueError("IDIFFR must be 1, 2, or 3")
        return v

    @field_validator("nqsiter")
    @classmethod
    def check_nqsiter(cls, v):
        if v < 1:
            raise ValueError("NQSITER must be at least 1")
        return v

    @field_validator("qsconv1")
    @classmethod
    def check_qsconv1(cls, v):
        if not 0 <= v <= 1:
            raise ValueError("QSCONV1 must be between 0 and 1")
        return v

    @field_validator("qsconv2")
    @classmethod
    def check_qsconv2(cls, v):
        if not 0 <= v <= 1:
            raise ValueError("QSCONV2 must be between 0 and 1")
        return v

    @field_validator("qsconv3")
    @classmethod
    def check_qsconv3(cls, v):
        if not 0 <= v <= 1:
            raise ValueError("QSCONV3 must be between 0 and 1")
        return v

    @field_validator("qsconv4")
    @classmethod
    def check_qsconv4(cls, v):
        if not 0 <= v <= 1:
            raise ValueError("QSCONV4 must be between 0 and 1")
        return v

    @field_validator("qsconv5")
    @classmethod
    def check_qsconv5(cls, v):
        if not 0 <= v <= 1:
            raise ValueError("QSCONV5 must be between 0 and 1")
        return v

    @field_validator("freqexp")
    @classmethod
    def check_freqexp(cls, v):
        if v <= 0:
            raise ValueError("FREQEXP must be positive")
        return v

    @field_validator("epsh1")
    @classmethod
    def check_epsh1(cls, v):
        if v <= 0:
            raise ValueError("EPSH1 must be positive")
        return v

    @field_validator("epsh2")
    @classmethod
    def check_epsh2(cls, v):
        if v <= 0:
            raise ValueError("EPSH2 must be positive")
        return v

    @field_validator("epsh3")
    @classmethod
    def check_epsh3(cls, v):
        if v <= 0:
            raise ValueError("EPSH3 must be positive")
        return v

    @field_validator("epsh4")
    @classmethod
    def check_epsh4(cls, v):
        if v <= 0:
            raise ValueError("EPSH4 must be positive")
        return v

    @field_validator("epsh5")
    @classmethod
    def check_epsh5(cls, v):
        if v <= 0:
            raise ValueError("EPSH5 must be positive")
        return v

    @field_validator("ivector")
    @classmethod
    def check_ivector(cls, v):
        if v not in range(1, 7):
            raise ValueError("IVECTOR must be between 1 and 6")
        return v

    @field_validator("dtmin_dyn")
    @classmethod
    def check_dtmin_dyn(cls, v):
        if v <= 0:
            raise ValueError("DTMIN_DYN must be positive")
        return v

    @field_validator("ndyniter")
    @classmethod
    def check_ndyniter(cls, v):
        if v < 1:
            raise ValueError("NDYNITER must be at least 1")
        return v

    @field_validator("dtmin_sin")
    @classmethod
    def check_dtmin_sin(cls, v):
        if v <= 0:
            raise ValueError("DTMIN_SIN must be positive")
        return v

    @field_validator("dtmin_snl4")
    @classmethod
    def check_dtmin_snl4(cls, v):
        if v <= 0:
            raise ValueError("DTMIN_SNL4 must be positive")
        return v

    @field_validator("dtmin_sds")
    @classmethod
    def check_dtmin_sds(cls, v):
        if v <= 0:
            raise ValueError("DTMIN_SDS must be positive")
        return v

    @field_validator("dtmin_snl3")
    @classmethod
    def check_dtmin_snl3(cls, v):
        if v <= 0:
            raise ValueError("DTMIN_SNL3 must be positive")
        return v

    @field_validator("dtmin_sbr")
    @classmethod
    def check_dtmin_sbr(cls, v):
        if v <= 0:
            raise ValueError("DTMIN_SBR must be positive")
        return v

    @field_validator("dtmin_sbf")
    @classmethod
    def check_dtmin_sbf(cls, v):
        if v <= 0:
            raise ValueError("DTMIN_SBF must be positive")
        return v

    @field_validator("ndyniter_sin")
    @classmethod
    def check_ndyniter_sin(cls, v):
        if v < 1:
            raise ValueError("NDYNITER_SIN must be at least 1")
        return v

    @field_validator("ndyniter_snl4")
    @classmethod
    def check_ndyniter_snl4(cls, v):
        if v < 1:
            raise ValueError("NDYNITER_SNL4 must be at least 1")
        return v

    @field_validator("ndyniter_sds")
    @classmethod
    def check_ndyniter_sds(cls, v):
        if v < 1:
            raise ValueError("NDYNITER_SDS must be at least 1")
        return v

    @field_validator("ndyniter_sbr")
    @classmethod
    def check_ndyniter_sbr(cls, v):
        if v < 1:
            raise ValueError("NDYNITER_SBR must be at least 1")
        return v

    @field_validator("ndyniter_snl3")
    @classmethod
    def check_ndyniter_snl3(cls, v):
        if v < 1:
            raise ValueError("NDYNITER_SNL3 must be at least 1")
        return v

    @field_validator("ndyniter_sbf")
    @classmethod
    def check_ndyniter_sbf(cls, v):
        if v < 1:
            raise ValueError("NDYNITER_SBF must be at least 1")
        return v

    @field_validator("wae_solverthr")
    @classmethod
    def check_wae_solverthr(cls, v):
        if v <= 0:
            raise ValueError("WAE_SOLVERTHR must be positive")
        return v

    @field_validator("maxiter")
    @classmethod
    def check_maxiter(cls, v):
        if v < 1:
            raise ValueError("MAXITER must be at least 1")
        return v

    @field_validator("pmin")
    @classmethod
    def check_pmin(cls, v):
        if not 0 <= v <= 100:
            raise ValueError("PMIN must be between 0 and 100")
        return v

    @field_validator("zeta_meth")
    @classmethod
    def check_zeta_meth(cls, v):
        if v < 0:
            raise ValueError("ZETA_METH must be non-negative")
        return v

    @field_validator("aspar_local_level")
    @classmethod
    def check_aspar_local_level(cls, v):
        if v not in range(6):
            raise ValueError("ASPAR_LOCAL_LEVEL must be between 0 and 5")
        return v


class History(NamelistBaseModel):
    begtc: Optional[str] = Field(
        "20200101.000000",
        description="Start output time in 'yyyymmdd.hhmmss' format. Must fit within the simulation time, otherwise no output. Defaults to PROC%BEGTC if not specified.",
    )
    deltc: Optional[int] = Field(
        3600,
        description="Time step for output in seconds. If smaller than simulation time step, the latter is used. Used for better 1D and 2D spectra analysis.",
    )
    unitc: Optional[str] = Field(
        "SEC",
        description="Unit of time for DELTC. Currently only supports 'SEC' for seconds.",
    )
    endtc: Optional[str] = Field(
        "20200201.000000",
        description="Stop time for output in 'yyyymmdd.hhmmss' format. Defaults to PROC%ENDC if not specified.",
    )
    definetc: Optional[int] = Field(
        -1,
        description="Time scoop (in seconds) for history files. If negative or unset, only one file is generated. For example, 86400 creates daily output files.",
    )
    outstyle: Optional[str] = Field(
        "NO",
        description="Output option. 'NO' for no output, 'NC' for netCDF, 'XFN' for XFN (default), 'SHP' for DARKO SHP output.",
    )
    multipleout: Optional[int] = Field(
        0,
        description="Output file configuration. 0 for single netCDF file using MPI_reduce (default), 1 for separate netCDF files for each process.",
    )
    use_single_out: Optional[bool] = Field(
        True,
        description="Use single precision in the output of model variables. True by default.",
    )
    paramwrite: Optional[bool] = Field(
        True,
        description="Write the physical parametrization and chosen numerical method in the netCDF file. True by default.",
    )
    gridwrite: Optional[bool] = Field(
        True, description="Write the grid in the netCDF history file. True by default."
    )
    printmma: Optional[bool] = Field(
        False,
        description="Print minimum, maximum and average value of statistics during runtime. Requires MPI_REDUCE. False by default.",
    )
    fileout: Optional[str] = Field(
        "wwm_hist.nc", description="Name of the output file."
    )
    hs: Optional[bool] = Field(True, description="Output significant wave height.")
    tm01: Optional[bool] = Field(True, description="Output mean period.")
    tm02: Optional[bool] = Field(True, description="Output zero-crossing mean period.")
    klm: Optional[bool] = Field(False, description="Output mean wave number.")
    wlm: Optional[bool] = Field(False, description="Output mean wave length.")
    etotc: Optional[bool] = Field(False, description="Output variable ETOTC.")
    etots: Optional[bool] = Field(False, description="Output variable ETOTS.")
    dm: Optional[bool] = Field(True, description="Output mean wave direction.")
    dspr: Optional[bool] = Field(True, description="Output directional spreading.")
    tppd: Optional[bool] = Field(
        False,
        description="Output direction of the peak (check source code for details).",
    )
    tpp: Optional[bool] = Field(True, description="Output peak period.")
    cpp: Optional[bool] = Field(False, description="Output peak phase velocity.")
    wnpp: Optional[bool] = Field(False, description="Output peak wave number.")
    cgpp: Optional[bool] = Field(False, description="Output peak group speed.")
    kpp: Optional[bool] = Field(False, description="Output peak wave number.")
    lpp: Optional[bool] = Field(False, description="Output peak wave length.")
    peakd: Optional[bool] = Field(True, description="Output peak direction.")
    peakdspr: Optional[bool] = Field(
        True, description="Output peak directional spreading."
    )
    dpeak: Optional[bool] = Field(False, description="Output peak direction.")
    ubot: Optional[bool] = Field(False, description="Output bottom excursion velocity.")
    orbital: Optional[bool] = Field(
        False, description="Output bottom orbital velocity."
    )
    botexper: Optional[bool] = Field(False, description="Output bottom excursion.")
    tmbot: Optional[bool] = Field(False, description="Output bottom period.")
    ursell: Optional[bool] = Field(False, description="Output Ursell number.")
    ufric: Optional[bool] = Field(False, description="Output air friction velocity.")
    z0: Optional[bool] = Field(False, description="Output air roughness length.")
    alpha_ch: Optional[bool] = Field(
        False, description="Output Charnock coefficient for air."
    )
    windx: Optional[bool] = Field(False, description="Output wind in X direction.")
    windy: Optional[bool] = Field(False, description="Output wind in Y direction.")
    cd: Optional[bool] = Field(False, description="Output drag coefficient.")
    currtx: Optional[bool] = Field(False, description="Output current in X direction.")
    currty: Optional[bool] = Field(False, description="Output current in Y direction.")
    watlev: Optional[bool] = Field(False, description="Output water level.")
    watlevold: Optional[bool] = Field(
        False, description="Output water level at previous time step."
    )
    depdt: Optional[bool] = Field(
        False, description="Output change of water level in time."
    )
    dep: Optional[bool] = Field(False, description="Output depth.")
    tauw: Optional[bool] = Field(
        True, description="Output surface stress from the wave."
    )
    tauhf: Optional[bool] = Field(
        False, description="Output high frequency surface stress."
    )
    tautot: Optional[bool] = Field(True, description="Output total surface stress.")
    stokessurfx: Optional[bool] = Field(
        True, description="Output surface Stokes drift in X direction."
    )
    stokessurfy: Optional[bool] = Field(
        True, description="Output surface Stokes drift in Y direction."
    )
    stokesbarox: Optional[bool] = Field(
        False, description="Output barotropic Stokes drift in X direction."
    )
    stokesbaroy: Optional[bool] = Field(
        False, description="Output barotropic Stokes drift in Y direction."
    )
    rsxx: Optional[bool] = Field(False, description="Output RSXX potential of LH.")
    rsxy: Optional[bool] = Field(False, description="Output RSXY potential of LH.")
    rsyy: Optional[bool] = Field(False, description="Output RSYY potential of LH.")
    cfl1: Optional[bool] = Field(False, description="Output CFL number 1.")
    cfl2: Optional[bool] = Field(False, description="Output CFL number 2.")
    cfl3: Optional[bool] = Field(False, description="Output CFL number 3.")

    @field_validator("begtc")
    @classmethod
    def validate_begtc(cls, v):
        try:
            datetime.strptime(v, "%Y%m%d.%H%M%S")
            return v
        except ValueError:
            raise ValueError("Invalid date format. Use yyyymmdd.hhmmss")

    @field_validator("deltc")
    @classmethod
    def validate_deltc(cls, v):
        if v <= 0:
            raise ValueError("DELTC must be positive")
        return v

    @field_validator("unitc")
    @classmethod
    def validate_unitc(cls, v):
        if v.upper() != "SEC":
            raise ValueError("UNITC must be SEC")
        return v.upper()

    @field_validator("endtc")
    @classmethod
    def validate_endtc(cls, v):
        try:
            datetime.strptime(v, "%Y%m%d.%H%M%S")
            return v
        except ValueError:
            raise ValueError("Invalid date format. Use yyyymmdd.hhmmss")

    @field_validator("definetc")
    @classmethod
    def validate_definetc(cls, v):
        if not isinstance(v, int):
            raise ValueError("DEFINETC must be an integer")
        return v

    @field_validator("outstyle")
    @classmethod
    def validate_outstyle(cls, v):
        valid_options = ["NO", "NC", "XFN", "SHP"]
        if v.upper() not in valid_options:
            raise ValueError(f"OUTSTYLE must be one of {valid_options}")
        return v.upper()

    @field_validator("multipleout")
    @classmethod
    def validate_multipleout(cls, v):
        if v not in [0, 1]:
            raise ValueError("MULTIPLEOUT must be 0 or 1")
        return v

    @field_validator("use_single_out")
    @classmethod
    def validate_use_single_out(cls, v):
        return v

    @field_validator("paramwrite")
    @classmethod
    def validate_paramwrite(cls, v):
        return v

    @field_validator("gridwrite")
    @classmethod
    def validate_gridwrite(cls, v):
        return v

    @field_validator("printmma")
    @classmethod
    def validate_printmma(cls, v):
        return v

    @field_validator("fileout")
    @classmethod
    def validate_fileout(cls, v):
        if not v.endswith(".nc"):
            raise ValueError("FILEOUT must have .nc extension")
        return v

    @field_validator("hs")
    @classmethod
    def validate_hs(cls, v):
        return v

    @field_validator("tm01")
    @classmethod
    def validate_tm01(cls, v):
        return v

    @field_validator("tm02")
    @classmethod
    def validate_tm02(cls, v):
        return v

    @field_validator("klm")
    @classmethod
    def validate_klm(cls, v):
        return v

    @field_validator("wlm")
    @classmethod
    def validate_wlm(cls, v):
        return v

    @field_validator("etotc")
    @classmethod
    def validate_etotc(cls, v):
        return v

    @field_validator("etots")
    @classmethod
    def validate_etots(cls, v):
        return v

    @field_validator("dm")
    @classmethod
    def validate_dm(cls, v):
        return v

    @field_validator("dspr")
    @classmethod
    def validate_dspr(cls, v):
        return v

    @field_validator("tppd")
    @classmethod
    def validate_tppd(cls, v):
        return v

    @field_validator("tpp")
    @classmethod
    def validate_tpp(cls, v):
        return v

    @field_validator("cpp")
    @classmethod
    def validate_cpp(cls, v):
        return v

    @field_validator("wnpp")
    @classmethod
    def validate_wnpp(cls, v):
        return v

    @field_validator("cgpp")
    @classmethod
    def validate_cgpp(cls, v):
        return v

    @field_validator("kpp")
    @classmethod
    def validate_kpp(cls, v):
        return v

    @field_validator("lpp")
    @classmethod
    def validate_lpp(cls, v):
        return v

    @field_validator("peakd")
    @classmethod
    def validate_peakd(cls, v):
        return v

    @field_validator("peakdspr")
    @classmethod
    def validate_peakdspr(cls, v):
        return v

    @field_validator("dpeak")
    @classmethod
    def validate_dpeak(cls, v):
        return v

    @field_validator("ubot")
    @classmethod
    def validate_ubot(cls, v):
        return v

    @field_validator("orbital")
    @classmethod
    def validate_orbital(cls, v):
        return v

    @field_validator("botexper")
    @classmethod
    def validate_botexper(cls, v):
        return v

    @field_validator("tmbot")
    @classmethod
    def validate_tmbot(cls, v):
        return v

    @field_validator("ursell")
    @classmethod
    def validate_ursell(cls, v):
        return v

    @field_validator("ufric")
    @classmethod
    def validate_ufric(cls, v):
        return v

    @field_validator("z0")
    @classmethod
    def validate_z0(cls, v):
        return v

    @field_validator("alpha_ch")
    @classmethod
    def validate_alpha_ch(cls, v):
        return v

    @field_validator("windx")
    @classmethod
    def validate_windx(cls, v):
        return v

    @field_validator("windy")
    @classmethod
    def validate_windy(cls, v):
        return v

    @field_validator("cd")
    @classmethod
    def validate_cd(cls, v):
        return v

    @field_validator("currtx")
    @classmethod
    def validate_currtx(cls, v):
        return v

    @field_validator("currty")
    @classmethod
    def validate_currty(cls, v):
        return v

    @field_validator("watlev")
    @classmethod
    def validate_watlev(cls, v):
        return v

    @field_validator("watlevold")
    @classmethod
    def validate_watlevold(cls, v):
        return v

    @field_validator("depdt")
    @classmethod
    def validate_depdt(cls, v):
        return v

    @field_validator("dep")
    @classmethod
    def validate_dep(cls, v):
        return v

    @field_validator("tauw")
    @classmethod
    def validate_tauw(cls, v):
        return v

    @field_validator("tauhf")
    @classmethod
    def validate_tauhf(cls, v):
        return v

    @field_validator("tautot")
    @classmethod
    def validate_tautot(cls, v):
        return v

    @field_validator("stokessurfx")
    @classmethod
    def validate_stokessurfx(cls, v):
        return v

    @field_validator("stokessurfy")
    @classmethod
    def validate_stokessurfy(cls, v):
        return v

    @field_validator("stokesbarox")
    @classmethod
    def validate_stokesbarox(cls, v):
        return v

    @field_validator("stokesbaroy")
    @classmethod
    def validate_stokesbaroy(cls, v):
        return v

    @field_validator("rsxx")
    @classmethod
    def validate_rsxx(cls, v):
        return v

    @field_validator("rsxy")
    @classmethod
    def validate_rsxy(cls, v):
        return v

    @field_validator("rsyy")
    @classmethod
    def validate_rsyy(cls, v):
        return v

    @field_validator("cfl1")
    @classmethod
    def validate_cfl1(cls, v):
        return v

    @field_validator("cfl2")
    @classmethod
    def validate_cfl2(cls, v):
        return v

    @field_validator("cfl3")
    @classmethod
    def validate_cfl3(cls, v):
        return v


class Station(NamelistBaseModel):
    begtc: Optional[str] = Field(
        "20200101.000000",
        description="Start simulation time in 'yyyymmdd.hhmmss' format. Must fit the simulation time, otherwise no output is generated. Defaults to PROC%BEGTC if not specified.",
    )
    deltc: Optional[int] = Field(
        3600,
        description="Time step for output in seconds. If smaller than simulation time step, the latter is used. Used for better 1D and 2D spectra analysis.",
    )
    unitc: Optional[str] = Field(
        "SEC", description="Time unit for DELTC. Only 'SEC' is currently supported."
    )
    endtc: Optional[str] = Field(
        "20200201.000000",
        description="Stop time for simulation in 'yyyymmdd.hhmmss' format. Defaults to PROC%ENDC if not specified.",
    )
    definetc: Optional[int] = Field(
        -1,
        description="Time for definition of station files in seconds. If negative or unset, only one file is generated. Otherwise, it defines the interval for creating output files (e.g., 86400 for daily files).",
    )
    outstyle: Optional[str] = Field(
        "NC",
        description="Output option. 'NO' for no output, 'STE' for classic station output (default), 'NC' for netCDF output.",
    )
    multipleout: Optional[int] = Field(
        0,
        description="Output file configuration. 0 for a single netCDF file using MPI_reduce (default), 1 for separate netCDF files for each process.",
    )
    use_single_out: Optional[bool] = Field(
        True,
        description="Use single precision in the output of model variables. True by default.",
    )
    paramwrite: Optional[bool] = Field(
        True,
        description="Write the physical parametrization and chosen numerical method in the netCDF file. True by default.",
    )
    fileout: Optional[str] = Field(
        "wwm_sta.nc", description="Output file name (not used)."
    )
    loutiter: Optional[bool] = Field(
        False, description="Boolean flag for output iteration."
    )
    iouts: Optional[int] = Field(12, description="Number of output stations.")
    nouts: Optional[list[str]] = Field(
        [
            "BatemansBay",
            "Brisbane",
            "ByronBay",
            "Cairns",
            "CoffsHbr",
            "CrowdyHead",
            "Eden",
            "HayPt",
            "Mackay",
            "PortKembla",
            "Tsv",
            "Mandurah",
        ],
        description="Names of output stations.",
    )
    xouts: Optional[list] = Field(
        [
            150.31972,
            153.63166,
            153.745,
            145.715167,
            153.27722,
            152.85333,
            150.15833,
            149.31025,
            149.5467,
            151.01667,
            147.059333,
            115.572227,
        ],
        description="Longitude coordinates of output stations.",
    )
    youts: Optional[list] = Field(
        [
            -35.75528,
            -27.48716,
            -28.67167,
            -16.7305,
            -30.34361,
            -31.82694,
            -37.175,
            -21.2715,
            -21.037333,
            -34.46694,
            -19.159167,
            -32.452787,
        ],
        description="Latitude coordinates of output stations.",
    )
    cutoff: Optional[float] = Field(
        0.0,
        description="Cutoff frequency (Hz) for each station, consistent with buoys.",
    )
    lsp1d: Optional[bool] = Field(
        True, description="Enable 1D spectral station output."
    )
    lsp2d: Optional[bool] = Field(
        True, description="Enable 2D spectral station output."
    )
    lsigmax: Optional[bool] = Field(
        True,
        description="Adjust the cut-off frequency for the output (e.g., consistent with buoy cut-off frequency).",
    )
    ac: Optional[bool] = Field(True, description="Output spectrum.")
    wk: Optional[bool] = Field(False, description="Output variable WK.")
    acout_1d: Optional[bool] = Field(False, description="Output variable ACOUT_1D.")
    acout_2d: Optional[bool] = Field(False, description="Output variable ACOUT_2D.")
    hs: Optional[bool] = Field(True, description="Output significant wave height.")
    tm01: Optional[bool] = Field(True, description="Output mean period.")
    tm02: Optional[bool] = Field(True, description="Output zero-crossing mean period.")
    klm: Optional[bool] = Field(False, description="Output mean wave number.")
    wlm: Optional[bool] = Field(False, description="Output mean wave length.")
    etotc: Optional[bool] = Field(False, description="Output variable ETOTC.")
    etots: Optional[bool] = Field(False, description="Output variable ETOTS.")
    dm: Optional[bool] = Field(True, description="Output mean wave direction.")
    dspr: Optional[bool] = Field(True, description="Output directional spreading.")
    tppd: Optional[bool] = Field(True, description="Output discrete peak period.")
    tpp: Optional[bool] = Field(True, description="Output peak period.")
    cpp: Optional[bool] = Field(False, description="Output variable CPP.")
    wnpp: Optional[bool] = Field(False, description="Output peak wave number.")
    cgpp: Optional[bool] = Field(False, description="Output peak group speed.")
    kpp: Optional[bool] = Field(False, description="Output peak wave number.")
    lpp: Optional[bool] = Field(False, description="Output peak wavelength.")
    peakd: Optional[bool] = Field(True, description="Output peak direction.")
    peakdspr: Optional[bool] = Field(
        True, description="Output peak directional spreading."
    )
    dpeak: Optional[bool] = Field(False, description="Output variable DPEAK.")
    ubot: Optional[bool] = Field(False, description="Output variable UBOT.")
    orbital: Optional[bool] = Field(False, description="Output orbital velocity.")
    botexper: Optional[bool] = Field(
        False, description="Output bottom excursion period."
    )
    tmbot: Optional[bool] = Field(False, description="Output variable TMBOT.")
    ursell: Optional[bool] = Field(False, description="Output Ursell number.")
    ufric: Optional[bool] = Field(False, description="Output air friction velocity.")
    z0: Optional[bool] = Field(False, description="Output air roughness length.")
    alpha_ch: Optional[bool] = Field(
        False, description="Output Charnock coefficient for air."
    )
    windx: Optional[bool] = Field(True, description="Output wind in X direction.")
    windy: Optional[bool] = Field(True, description="Output wind in Y direction.")
    cd: Optional[bool] = Field(False, description="Output drag coefficient.")
    currtx: Optional[bool] = Field(False, description="Output current in X direction.")
    currty: Optional[bool] = Field(False, description="Output current in Y direction.")
    watlev: Optional[bool] = Field(False, description="Output water level.")
    watlevold: Optional[bool] = Field(
        False, description="Output water level at previous time step."
    )
    depdt: Optional[bool] = Field(
        False, description="Output change of water level in time."
    )
    dep: Optional[bool] = Field(True, description="Output depth.")
    tauw: Optional[bool] = Field(
        False, description="Output surface stress from the wave."
    )
    tauhf: Optional[bool] = Field(
        False, description="Output high frequency surface stress."
    )
    tautot: Optional[bool] = Field(False, description="Output total surface stress.")
    stokessurfx: Optional[bool] = Field(
        True, description="Output surface Stokes drift in X direction."
    )
    stokessurfy: Optional[bool] = Field(
        True, description="Output surface Stokes drift in Y direction."
    )
    stokesbarox: Optional[bool] = Field(
        False, description="Output barotropic Stokes drift in X direction."
    )
    stokesbaroy: Optional[bool] = Field(
        False, description="Output barotropic Stokes drift in Y direction."
    )
    rsxx: Optional[bool] = Field(False, description="Output RSXX potential of LH.")
    rsxy: Optional[bool] = Field(False, description="Output RSXY potential of LH.")
    rsyy: Optional[bool] = Field(False, description="Output RSYY potential of LH.")
    cfl1: Optional[bool] = Field(False, description="Output CFL number 1.")
    cfl2: Optional[bool] = Field(False, description="Output CFL number 2.")
    cfl3: Optional[bool] = Field(False, description="Output CFL number 3.")

    @field_validator("begtc")
    def validate_begtc(cls, v):
        import re

        if not re.match(r"^\d{8}\.\d{6}$", v):
            raise ValueError("BEGTC must be in format yyyymmdd.hhmmss")
        return v

    @field_validator("deltc")
    def validate_deltc(cls, v):
        if v <= 0:
            raise ValueError("DELTC must be a positive integer")
        return v

    @field_validator("unitc")
    def validate_unitc(cls, v):
        if v != "SEC":
            raise ValueError("UNITC must be SEC")
        return v

    @field_validator("endtc")
    def validate_endtc(cls, v):
        import re

        if not re.match(r"^\d{8}\.\d{6}$", v):
            raise ValueError("ENDTC must be in format yyyymmdd.hhmmss")
        return v

    @field_validator("definetc")
    def validate_definetc(cls, v):
        if v != -1 and v < 0:
            raise ValueError("DEFINETC must be -1 or a non-negative integer")
        return v

    @field_validator("outstyle")
    def validate_outstyle(cls, v):
        if v not in ["NO", "STE", "NC"]:
            raise ValueError("OUTSTYLE must be NO, STE, or NC")
        return v

    @field_validator("multipleout")
    def validate_multipleout(cls, v):
        if v not in [0, 1]:
            raise ValueError("MULTIPLEOUT must be 0 or 1")
        return v

    @field_validator("iouts")
    def validate_iouts(cls, v):
        if v <= 0:
            raise ValueError("IOUTS must be a positive integer")
        return v

    @field_validator("nouts")
    def validate_nouts(cls, v, info):
        if len(v) != info.data.get("iouts"):
            raise ValueError("Number of NOUTS must match IOUTS")
        return v

    @field_validator("xouts")
    def validate_xouts(cls, v, info):
        if len(v) != info.data.get("iouts"):
            raise ValueError("Number of XOUTS must match IOUTS")
        return v

    @field_validator("youts")
    def validate_youts(cls, v, info):
        if len(v) != info.data.get("iouts"):
            raise ValueError("Number of YOUTS must match IOUTS")
        return v

    @field_validator("cutoff")
    def validate_cutoff(cls, v):
        if v < 0:
            raise ValueError("CUTOFF must be non-negative")
        return v


class Hotfile(NamelistBaseModel):
    lhotf: Optional[bool] = Field(True, description="Write hotfile")
    filehot_out: Optional[str] = Field("wwm_hot_out.nc", description="name of output")
    begtc: Optional[str] = Field(
        "20200101.000000",
        description="Starting time of hotfile writing. With ihot!=0 in SCHISM,",
    )
    deltc: Optional[float] = Field(2678400.0, description="time between hotfile writes")
    unitc: Optional[str] = Field("SEC", description="unit used above")
    endtc: Optional[str] = Field(
        "20200201.000000",
        description="Ending time of hotfile writing (adjust with BEGTC)",
    )
    lcyclehot: Optional[bool] = Field(False, description="Applies only to netcdf")
    hotstyle_out: Optional[int] = Field(
        2, description="1: binary hotfile of data as output"
    )
    multipleout: Optional[int] = Field(
        0, description="0: hotfile in a single file (binary or netcdf)"
    )
    filehot_in: Optional[str] = Field(
        "wwm_hot_in.nc",
        description="(Full) hot file name for input (which can be copied from FILEHOT_OUT above)",
    )
    hotstyle_in: Optional[int] = Field(
        2, description="1: binary hotfile of data as input"
    )
    ihotpos_in: Optional[int] = Field(
        2, description="Position in hotfile (only for netcdf)"
    )
    multiplein: Optional[int] = Field(
        0, description="0: read hotfile from one single file"
    )


class Petscoptions(NamelistBaseModel):
    ksptype: Optional[str] = Field(
        "LGMRES",
        description="Controls the linear solver type used by PETSc. Options include GMRES, LGMRES (augmented GMRES), DGMRES (deflated GMRES), PGMRES (pipelined GMRES), and KSPBCGSL (variant of Enhanced BiCGStab(L)).",
    )
    rtol: Optional[float] = Field(
        1e-20,
        description="Relative convergence tolerance, representing the relative decrease in the residual norm for the iterative solver.",
    )
    abstol: Optional[float] = Field(
        1e-20,
        description="Absolute convergence tolerance, representing the absolute size of the residual norm for the iterative solver.",
    )
    dtol: Optional[float] = Field(
        10000.0, description="Divergence tolerance for the iterative solver."
    )
    maxits: Optional[int] = Field(
        1000,
        description="Maximum number of iterations allowed for the iterative solver.",
    )
    initialguessnonzero: Optional[bool] = Field(
        False,
        description="Boolean flag indicating whether the initial guess for the iterative solver is nonzero.",
    )
    gmrespreallocate: Optional[bool] = Field(
        True,
        description="Boolean flag indicating whether GMRES and FGMRES should preallocate all needed work vectors at initial setup.",
    )
    pctype: Optional[str] = Field(
        "SOR",
        description="Controls the preconditioner type used by PETSc. Options include SOR (successive over relaxation), ASM (additive Schwarz method), HYPRE (LLNL package), SPAI (Sparse Approximate Inverse), and NONE (no preconditioning).",
    )

    @field_validator("ksptype")
    @classmethod
    def validate_ksptype(cls, v):
        valid_types = ["GMRES", "LGMRES", "DGMRES", "PGMRES", "KSPBCGSL"]
        if v.upper() not in valid_types:
            raise ValueError(f"KSPTYPE must be one of {valid_types}")
        return v.upper()

    @field_validator("rtol")
    @classmethod
    def validate_rtol(cls, v):
        if v <= 0:
            raise ValueError("RTOL must be positive")
        return v

    @field_validator("abstol")
    @classmethod
    def validate_abstol(cls, v):
        if v <= 0:
            raise ValueError("ABSTOL must be positive")
        return v

    @field_validator("dtol")
    @classmethod
    def validate_dtol(cls, v):
        if v <= 0:
            raise ValueError("DTOL must be positive")
        return v

    @field_validator("maxits")
    @classmethod
    def validate_maxits(cls, v):
        if v <= 0 or not isinstance(v, int):
            raise ValueError("MAXITS must be a positive integer")
        return v

    @field_validator("initialguessnonzero")
    @classmethod
    def validate_initialguessnonzero(cls, v):
        return v

    @field_validator("gmrespreallocate")
    @classmethod
    def validate_gmrespreallocate(cls, v):
        return v

    @field_validator("pctype")
    @classmethod
    def validate_pctype(cls, v):
        valid_types = ["SOR", "ASM", "HYPRE", "SPAI", "NONE"]
        if v.upper() not in valid_types:
            raise ValueError(f"PCTYPE must be one of {valid_types}")
        return v.upper()


# class Nesting(NamelistBaseModel):
#     listbegtc: Optional[str] = Field(
#         "", description="Start date/time for the nested grid simulation"
#     )
#     listdeltc: Optional[str] = Field(
#         "ZERO", description="Time step for the nested grid simulation"
#     )
#     listunitc: Optional[str] = Field(
#         "", description="Time unit for the nested grid simulation"
#     )
#     listendtc: Optional[str] = Field(
#         "", description="End date/time for the nested grid simulation"
#     )
#     listigridtype: Optional[int] = Field(
#         0,
#         description="Grid type for the nested simulation (0 for regular lat-lon grid)",
#     )
#     listfilegrid: Optional[str] = Field(
#         "", description="File containing grid definition for the nested simulation"
#     )
#     listfilebound: Optional[str] = Field(
#         "", description="File containing boundary conditions for the nested simulation"
#     )
#     listprefix: Optional[str] = Field(
#         "", description="Prefix for output files from the nested simulation"
#     )
#
# @field_validator("listbegtc")
# @classmethod
# def validate_listbegtc(cls, v):
#     if not v:
#         raise ValueError("ListBEGTC must not be empty")
#     return v
#
# @field_validator("listdeltc")
# @classmethod
# def validate_listdeltc(cls, v):
#     if v.upper() != "ZERO":
#         raise ValueError("ListDELTC must be ZERO")
#     return v
#
# @field_validator("listunitc")
# @classmethod
# def validate_listunitc(cls, v):
#     if not v:
#         raise ValueError("ListUNITC must not be empty")
#     return v
#
# @field_validator("listendtc")
# @classmethod
# def validate_listendtc(cls, v):
#     if not v:
#         raise ValueError("ListENDTC must not be empty")
#     return v
#
# @field_validator("listigridtype")
# @classmethod
# def validate_listigridtype(cls, v):
#     if v != 0:
#         raise ValueError("ListIGRIDTYPE must be 0")
#     return v
#
# @field_validator("listfilegrid")
# @classmethod
# def validate_listfilegrid(cls, v):
#     if not v:
#         raise ValueError("ListFILEGRID must not be empty")
#     return v
#
# @field_validator("listfilebound")
# @classmethod
# def validate_listfilebound(cls, v):
#     if not v:
#         raise ValueError("ListFILEBOUND must not be empty")
#     return v
#
# @field_validator("listprefix")
# @classmethod
# def validate_listprefix(cls, v):
#     if not v:
#         raise ValueError("ListPrefix must not be empty")
#     return v


class Wind(NamelistBaseModel):
    "Empty class to fill empty wind namelist"


class Curr(NamelistBaseModel):
    "Empty class to fill empty curr namelist"


class Walv(NamelistBaseModel):
    "Empty class to fill empty wavl namelist"


class Nesting(NamelistBaseModel):
    "Empty class to fill empty nesting namelist"


class Wwminput(NamelistBaseModel):
    proc: Optional[Proc] = Field(default_factory=Proc)
    coupl: Optional[Coupl] = Field(default_factory=Coupl)
    grid: Optional[Grid] = Field(default_factory=Grid)
    init: Optional[Init] = Field(default_factory=Init)
    bouc: Optional[Bouc] = Field(default_factory=Bouc)
    wind: Optional[Wind] = Field(default_factory=Wind)
    curr: Optional[Curr] = Field(default_factory=Curr)
    walv: Optional[Walv] = Field(default_factory=Walv)
    engs: Optional[Engs] = Field(default_factory=Engs)
    nums: Optional[Nums] = Field(default_factory=Nums)
    history: Optional[History] = Field(default_factory=History)
    station: Optional[Station] = Field(default_factory=Station)
    hotfile: Optional[Hotfile] = Field(default_factory=Hotfile)
    petscoptions: Optional[Petscoptions] = Field(default_factory=Petscoptions)
    nesting: Optional[Nesting] = Field(default_factory=Nesting)

    def render(self) -> str:
        # Needs an additional backslash due to to uknown issue with schism
        # All sameple namelists have this feature
        return super().render() + "\n" + "/"



"""
Boundary plotting extensions for SCHISMConfig.

This module contains methods for visualizing boundary data in SCHISM models.
"""


def plot_boundary_points(self, variable=None, ax=None, figsize=(10, 8), **kwargs):
    """Plot boundary node points from a boundary dataset.

    Parameters
    ----------
    variable : str, optional
        Variable to color the points by. If None, points are not colored by variable.
    ax : matplotlib.axes.Axes, optional
        The axes to plot on. If None, a new figure is created.
    figsize : tuple, optional
        Figure size for new figure. Default is (10, 8).
    **kwargs : dict
        Additional keyword arguments to pass to scatter plot.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object.
    ax : matplotlib.axes.Axes
        The axes object.
    """
    import logging

    import matplotlib.pyplot as plt

    logger = logging.getLogger(__name__)

    # Check if we have the necessary components
    if not hasattr(self, "grid") or self.grid is None:
        raise ValueError("No grid available in the config object")
    if not hasattr(self, "data") or self.data is None:
        raise ValueError("No data available in the config object")

    # Check for boundary data in the correct location (under ocean)
    boundary = None
    if (
        hasattr(self.data, "ocean")
        and self.data.ocean is not None
        and hasattr(self.data.ocean, "boundary")
        and self.data.ocean.boundary is not None
    ):
        boundary = self.data.ocean.boundary
    # Fall back to legacy location for backward compatibility
    elif hasattr(self.data, "boundary") and self.data.boundary is not None:
        boundary = self.data.boundary
    else:
        raise ValueError(
            "No boundary data available in the config object (checked both data.ocean.boundary and data.boundary)"
        )

    # Get the grid and boundary dataset
    grid = self.grid
    dataset = boundary.source.dataset

    # Create plot if needed
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.get_figure()

    # Plot the grid boundaries
    try:
        grid.plot_boundary(ax=ax, color="k", linewidth=1.0)
    except Exception as e:
        logger.warning(f"Could not plot grid boundaries: {e}")

    # Get boundary node indices
    boundary_nodes = []
    for i in range(grid.pylibs_hgrid.nob):
        boundary_nodes.extend(grid.pylibs_hgrid.iobn[i])

    # Get boundary node coordinates
    x = grid.pylibs_hgrid.x[boundary_nodes]
    y = grid.pylibs_hgrid.y[boundary_nodes]

    # Color by variable if specified
    if variable is not None and variable in dataset:
        # Get variable values for coloring
        if "node" in dataset.dims and len(dataset[variable].shape) >= 1:
            # Extract values for the first time step if available
            if "time" in dataset.dims and len(dataset[variable].shape) >= 2:
                values = dataset[variable].isel(time=0).values
            else:
                values = dataset[variable].values

            # Plot colored points
            scatter = ax.scatter(x, y, c=values, **kwargs)
            plt.colorbar(scatter, ax=ax, label=variable)
        else:
            logger.warning(
                f"Variable {variable} not suitable for coloring boundary points"
            )
            ax.scatter(x, y, **kwargs)
    else:
        # Plot without variable coloring
        ax.scatter(x, y, **kwargs)

    # Set labels and title
    ax.set_xlabel("Longitude")
    ax.set_ylabel("Latitude")
    ax.set_title("Boundary Points")
    ax.set_aspect("equal")

    return fig, ax


def plot_boundary_timeseries(
    self, variable=None, boundary_idx=0, node_idx=0, ax=None, figsize=(12, 6), **kwargs
):
    """Plot time series of a boundary variable at a specific boundary node.

    Parameters
    ----------
    variable : str, optional
        Variable to plot. If None, the first variable in the dataset is used.
    boundary_idx : int, optional
        Index of the boundary to plot. Default is 0 (first boundary).
    node_idx : int, optional
        Index of the node within the boundary to plot. Default is 0 (first node).
    ax : matplotlib.axes.Axes, optional
        The axes to plot on. If None, a new figure is created.
    figsize : tuple, optional
        Figure size for new figure. Default is (12, 6).
    **kwargs : dict
        Additional keyword arguments to pass to plot.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object.
    """
    import logging

    import matplotlib.pyplot as plt
    import numpy as np

    logger = logging.getLogger(__name__)

    # Check if we have the necessary components
    if not hasattr(self, "grid") or self.grid is None:
        raise ValueError("No grid available in the config object")
    if not hasattr(self, "data") or self.data is None:
        raise ValueError("No data available in the config object")

    # Check for boundary data in the correct location (under ocean)
    boundary = None
    if (
        hasattr(self.data, "ocean")
        and self.data.ocean is not None
        and hasattr(self.data.ocean, "boundary")
        and self.data.ocean.boundary is not None
    ):
        boundary = self.data.ocean.boundary
    # Fall back to legacy location for backward compatibility
    elif hasattr(self.data, "boundary") and self.data.boundary is not None:
        boundary = self.data.boundary
    else:
        raise ValueError(
            "No boundary data available in the config object (checked both data.ocean.boundary and data.boundary)"
        )

    # Get the grid and boundary dataset
    grid = self.grid
    dataset = boundary.source.dataset

    # Check if dataset has time dimension
    if "time" not in dataset.dims:
        raise ValueError("Dataset does not have a time dimension")

    # Get first variable if not specified
    if variable is None and len(dataset.data_vars) > 0:
        variable = list(dataset.data_vars)[0]

    if variable not in dataset:
        raise ValueError(f"Variable '{variable}' not found in dataset")

    # Create plot if needed
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.get_figure()

    # Get boundary node indices
    if boundary_idx >= grid.pylibs_hgrid.nob:
        raise ValueError(
            f"Boundary index {boundary_idx} out of range (0-{grid.pylibs_hgrid.nob-1})"
        )

    boundary_nodes = grid.pylibs_hgrid.iobn[boundary_idx]
    if node_idx >= len(boundary_nodes):
        raise ValueError(
            f"Node index {node_idx} out of range (0-{len(boundary_nodes)-1})"
        )

    # Get global node index
    global_node_idx = boundary_nodes[node_idx]

    # Get coordinates of the node
    x = grid.pylibs_hgrid.x[global_node_idx]
    y = grid.pylibs_hgrid.y[global_node_idx]

    # Extract time series
    times = dataset["time"].values

    # Find the appropriate dimension for nodes
    node_dim = None
    for dim in dataset[variable].dims:
        if dim in ["node", "nOpenBndNodes", "nBnd"]:
            node_dim = dim
            break

    if node_dim is None:
        raise ValueError(f"Could not find node dimension in variable {variable}")

    # Map global node index to dataset index if needed
    if "node" in dataset.coords:
        # Check if dataset nodes match grid nodes
        dataset_nodes = dataset["node"].values
        if global_node_idx in dataset_nodes:
            dataset_node_idx = np.where(dataset_nodes == global_node_idx)[0][0]
        else:
            # If not, just use the local node index within boundary
            logger.warning(
                "Node indices in dataset do not match grid node indices. Using local index."
            )
            dataset_node_idx = node_idx
    else:
        # If no node coordinate, assume nodes are in the same order as in the grid
        dataset_node_idx = node_idx

    # Extract values
    if len(dataset[variable].dims) == 2 and node_dim in dataset[variable].dims:
        # 2D variable (time, node)
        values = dataset[variable].isel({node_dim: dataset_node_idx}).values
    elif len(dataset[variable].dims) == 3 and node_dim in dataset[variable].dims:
        # 3D variable (time, node, level) - extract surface level
        values = dataset[variable].isel({node_dim: dataset_node_idx, "level": 0}).values
    else:
        raise ValueError(f"Variable {variable} does not have expected dimensions")

    # Plot the time series
    ax.plot(times, values, **kwargs)

    # Add labels and title
    ax.set_xlabel("Time")
    units = dataset[variable].attrs.get("units", "")
    ax.set_ylabel(f"{variable} ({units})" if units else variable)
    ax.set_title(
        f"{variable} at Boundary {boundary_idx}, Node {node_idx} ({x:.2f}, {y:.2f})"
    )

    # Format the x-axis to show dates nicely
    fig.autofmt_xdate()

    return fig


def plot_boundary_profile(
    self,
    variable=None,
    boundary_idx=0,
    node_idx=0,
    time_idx=0,
    ax=None,
    figsize=(8, 10),
    **kwargs,
):
    """Plot depth profile of a boundary variable at a specific boundary node and time.

    Parameters
    ----------
    variable : str, optional
        Variable to plot. If None, the first 3D variable in the dataset is used.
    boundary_idx : int, optional
        Index of the boundary to plot. Default is 0 (first boundary).
    node_idx : int, optional
        Index of the node within the boundary to plot. Default is 0 (first node).
    time_idx : int, optional
        Index of the time step to plot. Default is 0 (first time step).
    ax : matplotlib.axes.Axes, optional
        The axes to plot on. If None, a new figure is created.
    figsize : tuple, optional
        Figure size for new figure. Default is (8, 10).
    **kwargs : dict
        Additional keyword arguments to pass to plot.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object.
    """
    import logging

    import matplotlib.pyplot as plt
    import numpy as np

    logger = logging.getLogger(__name__)

    # Check if we have the necessary components
    if not hasattr(self, "grid") or self.grid is None:
        raise ValueError("No grid available in the config object")
    if not hasattr(self, "data") or self.data is None:
        raise ValueError("No data available in the config object")

    # Check for boundary data in the correct location (under ocean)
    boundary = None
    if (
        hasattr(self.data, "ocean")
        and self.data.ocean is not None
        and hasattr(self.data.ocean, "boundary")
        and self.data.ocean.boundary is not None
    ):
        boundary = self.data.ocean.boundary
    # Fall back to legacy location for backward compatibility
    elif hasattr(self.data, "boundary") and self.data.boundary is not None:
        boundary = self.data.boundary
    else:
        raise ValueError(
            "No boundary data available in the config object (checked both data.ocean.boundary and data.boundary)"
        )

    # Get the grid and boundary dataset
    grid = self.grid
    dataset = boundary.source.dataset

    # Check if dataset has depth/level dimension
    depth_dim = None
    for dim in dataset.dims:
        if dim in ["depth", "level", "nLevels", "sigma"]:
            depth_dim = dim
            break

    if depth_dim is None:
        raise ValueError("Dataset does not have a depth/level dimension")

    # Find a 3D variable if not specified
    if variable is None:
        for var in dataset.data_vars:
            if len(dataset[var].dims) >= 3 and depth_dim in dataset[var].dims:
                variable = var
                break

    if variable not in dataset or depth_dim not in dataset[variable].dims:
        raise ValueError(
            f"Variable '{variable}' not found in dataset or does not have a depth dimension"
        )

    # Create plot if needed
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.get_figure()

    # Get boundary node indices
    if boundary_idx >= grid.pylibs_hgrid.nob:
        raise ValueError(
            f"Boundary index {boundary_idx} out of range (0-{grid.pylibs_hgrid.nob-1})"
        )

    boundary_nodes = grid.pylibs_hgrid.iobn[boundary_idx]
    if node_idx >= len(boundary_nodes):
        raise ValueError(
            f"Node index {node_idx} out of range (0-{len(boundary_nodes)-1})"
        )

    # Get global node index
    global_node_idx = boundary_nodes[node_idx]

    # Get coordinates of the node
    x = grid.pylibs_hgrid.x[global_node_idx]
    y = grid.pylibs_hgrid.y[global_node_idx]

    # Find the appropriate dimension for nodes
    node_dim = None
    for dim in dataset[variable].dims:
        if dim in ["node", "nOpenBndNodes", "nBnd"]:
            node_dim = dim
            break

    if node_dim is None:
        raise ValueError(f"Could not find node dimension in variable {variable}")

    # Map global node index to dataset index if needed
    if "node" in dataset.coords:
        # Check if dataset nodes match grid nodes
        dataset_nodes = dataset["node"].values
        if global_node_idx in dataset_nodes:
            dataset_node_idx = np.where(dataset_nodes == global_node_idx)[0][0]
        else:
            # If not, just use the local node index within boundary
            logger.warning(
                "Node indices in dataset do not match grid node indices. Using local index."
            )
            dataset_node_idx = node_idx
    else:
        # If no node coordinate, assume nodes are in the same order as in the grid
        dataset_node_idx = node_idx

    # Get depth values
    depths = dataset[depth_dim].values

    # Extract profile values
    if "time" in dataset[variable].dims:
        values = (
            dataset[variable].isel(time=time_idx, **{node_dim: dataset_node_idx}).values
        )
    else:
        values = dataset[variable].isel({node_dim: dataset_node_idx}).values

    # Plot the depth profile
    ax.plot(values, depths, **kwargs)

    # Reverse y-axis to have surface at top
    ax.invert_yaxis()

    # Add labels and title
    units = dataset[variable].attrs.get("units", "")
    ax.set_xlabel(f"{variable} ({units})" if units else variable)
    ax.set_ylabel(f'Depth ({dataset[depth_dim].attrs.get("units", "m")})')

    # Add time to title if available
    title = f"{variable} Profile at Boundary {boundary_idx}, Node {node_idx} ({x:.2f}, {y:.2f})"
    if "time" in dataset:
        try:
            time_str = str(dataset["time"].values[time_idx])
            title += f"\nTime: {time_str}"
        except:
            pass

    ax.set_title(title)

    return fig



# This file was auto generated from a SCHISM namelist file on 2025-01-24.

from typing import Optional

from pydantic import Field, field_validator, model_validator

from rompy_schism.namelists.basemodel import NamelistBaseModel


class Ice_in(NamelistBaseModel):
    ice_tests: Optional[int] = Field(
        0, description="Flag for box test. 0 disables, 1 enables."
    )
    ice_advection: Optional[int] = Field(
        1, description="Flag to enable/disable ice advection. 1 enables, 0 disables."
    )
    ice_therm_on: Optional[int] = Field(
        1,
        description="Flag to enable/disable ice thermodynamics. 1 enables, 0 disables.",
    )
    ievp: Optional[int] = Field(
        2,
        description="Selects the rheology model. 1 for EVP (Elastic-Viscous-Plastic), 2 for mEVP (modified EVP).",
    )
    ice_cutoff: Optional[float] = Field(
        0.001,
        description="Cut-off thickness [m] or fraction for ice. No ice velocity if ice thickness is less than or equal to ice_cutoff.",
    )
    evp_rheol_steps: Optional[int] = Field(
        200, description="Number of subcycling steps in EVP rheology model."
    )
    mevp_rheol_steps: Optional[int] = Field(
        200, description="Number of iterations in mEVP rheology model."
    )
    ice_atmos_stress_form: Optional[int] = Field(
        1,
        description="Form of atmospheric stress calculation. 0 for constant Cd, 1 for FESOM formulation.",
    )
    cdwin0: Optional[float] = Field(
        0.002,
        description="Constant drag coefficient for wind stress, used if ice_atmos_stress_form=0.",
    )
    delta_min: Optional[float] = Field(
        2e-09,
        description="Limit for minimum divergence (1/s). Used in both VP and EVP rheology models.",
    )
    theta_io: Optional[float] = Field(
        0.0,
        description="Ice/ocean rotation angle in degrees. Usually 0 unless vertical grid is too coarse.",
    )
    mevp_coef: Optional[int] = Field(
        0,
        description="Options for specifying 2 relax coefficients in mEVP. 0 for constant, 1 for variable coefficients.",
    )
    mevp_alpha1: Optional[float] = Field(
        200.0,
        description="Constant used in mEVP for constitutive equation if mevp_coef=0.",
    )
    mevp_alpha2: Optional[float] = Field(
        200.0, description="Constant used in mEVP for momentum equation if mevp_coef=0."
    )
    mevp_alpha3: Optional[float] = Field(
        200.0, description="Minimum value for variable coefficients if mevp_coef=1."
    )
    mevp_alpha4: Optional[float] = Field(
        0.02,
        description="Coefficient used in variable coefficient calculation if mevp_coef=1.",
    )
    pstar: Optional[float] = Field(
        15000.0, description="Ice strength parameter [N/m^2]."
    )
    ellipse: Optional[float] = Field(2.0, description="Ellipticity of the yield curve.")
    c_pressure: Optional[float] = Field(
        20.0, description="Ice concentration parameter C [-]."
    )
    ncyc_fct: Optional[int] = Field(
        1, description="Number of subcycling steps in transport for FCT scheme."
    )
    niter_fct: Optional[int] = Field(
        3, description="Number of iterations in higher-order solve for FCT scheme."
    )
    ice_gamma_fct: Optional[float] = Field(
        0.25,
        description="Smoothing parameter for FCT scheme. 1 for maximum positivity preserving.",
    )
    depth_ice_fct: Optional[float] = Field(
        5.0, description="Cut-off depth (m) for non-FCT zone in ice_fct.gr3."
    )
    h_ml0: Optional[float] = Field(
        0.1, description="Ocean mixed layer depth [m] for thermodynamics calculations."
    )
    salt_ice: Optional[float] = Field(
        5.0, description="Salinity for ice [PSU] (must be non-negative)."
    )
    salt_water: Optional[float] = Field(
        34.0, description="Salinity for water [PSU] (must be non-negative)."
    )
    lead_closing: Optional[float] = Field(
        0.5,
        description="Lead closing parameter [m]. Larger values slow down freezing-up but increase sea ice thickness.",
    )
    saterm: Optional[float] = Field(
        0.5, description="Semter constant. Smaller values could slow down melting."
    )
    albsn: Optional[float] = Field(0.85, description="Albedo for frozen snow.")
    albsnm: Optional[float] = Field(
        0.75,
        description="Albedo for melting snow (must be less than or equal to albsn).",
    )
    albi: Optional[float] = Field(
        0.75, description="Albedo for frozen ice (must be less than or equal to albsn)."
    )
    albm: Optional[float] = Field(
        0.66, description="Albedo for melting ice (must be less than or equal to albi)."
    )

    @field_validator("ice_tests")
    @classmethod
    def validate_ice_tests(cls, v):
        if not isinstance(v, int) or v not in [0, 1]:
            raise ValueError("ice_tests must be 0 or 1")
        return v

    @field_validator("ice_advection")
    @classmethod
    def validate_ice_advection(cls, v):
        if not isinstance(v, int) or v not in [0, 1]:
            raise ValueError("ice_advection must be 0 or 1")
        return v

    @field_validator("ice_therm_on")
    @classmethod
    def validate_ice_therm_on(cls, v):
        if not isinstance(v, int) or v not in [0, 1]:
            raise ValueError("ice_therm_on must be 0 or 1")
        return v

    @field_validator("ievp")
    @classmethod
    def validate_ievp(cls, v):
        if not isinstance(v, int) or v not in [1, 2]:
            raise ValueError("ievp must be 1 or 2")
        return v

    @field_validator("ice_cutoff")
    @classmethod
    def validate_ice_cutoff(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("ice_cutoff must be a positive float")
        return v

    @field_validator("evp_rheol_steps")
    @classmethod
    def validate_evp_rheol_steps(cls, v):
        if not isinstance(v, int) or v <= 0:
            raise ValueError("evp_rheol_steps must be a positive integer")
        return v

    @field_validator("mevp_rheol_steps")
    @classmethod
    def validate_mevp_rheol_steps(cls, v):
        if not isinstance(v, int) or v <= 0:
            raise ValueError("mevp_rheol_steps must be a positive integer")
        return v

    @field_validator("ice_atmos_stress_form")
    @classmethod
    def validate_ice_atmos_stress_form(cls, v):
        if not isinstance(v, int) or v not in [0, 1]:
            raise ValueError("ice_atmos_stress_form must be 0 or 1")
        return v

    @field_validator("cdwin0")
    @classmethod
    def validate_cdwin0(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("cdwin0 must be a positive float")
        return v

    @field_validator("delta_min")
    @classmethod
    def validate_delta_min(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("delta_min must be a positive float")
        return v

    @field_validator("theta_io")
    @classmethod
    def validate_theta_io(cls, v):
        if not isinstance(v, float) or v < 0 or v >= 360:
            raise ValueError("theta_io must be a float between 0 and 360")
        return v

    @field_validator("mevp_coef")
    @classmethod
    def validate_mevp_coef(cls, v):
        if not isinstance(v, int) or v not in [0, 1]:
            raise ValueError("mevp_coef must be 0 or 1")
        return v

    @field_validator("mevp_alpha1")
    @classmethod
    def validate_mevp_alpha1(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("mevp_alpha1 must be a positive float")
        return v

    @field_validator("mevp_alpha2")
    @classmethod
    def validate_mevp_alpha2(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("mevp_alpha2 must be a positive float")
        return v

    @field_validator("mevp_alpha3")
    @classmethod
    def validate_mevp_alpha3(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("mevp_alpha3 must be a positive float")
        return v

    @field_validator("mevp_alpha4")
    @classmethod
    def validate_mevp_alpha4(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("mevp_alpha4 must be a positive float")
        return v

    @field_validator("pstar")
    @classmethod
    def validate_pstar(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("pstar must be a positive float")
        return v

    @field_validator("ellipse")
    @classmethod
    def validate_ellipse(cls, v):
        if not isinstance(v, float) or v <= 1:
            raise ValueError("ellipse must be a float greater than 1")
        return v

    @field_validator("c_pressure")
    @classmethod
    def validate_c_pressure(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("c_pressure must be a positive float")
        return v

    @field_validator("ncyc_fct")
    @classmethod
    def validate_ncyc_fct(cls, v):
        if not isinstance(v, int) or v <= 0:
            raise ValueError("ncyc_fct must be a positive integer")
        return v

    @field_validator("niter_fct")
    @classmethod
    def validate_niter_fct(cls, v):
        if not isinstance(v, int) or v <= 0:
            raise ValueError("niter_fct must be a positive integer")
        return v

    @field_validator("ice_gamma_fct")
    @classmethod
    def validate_ice_gamma_fct(cls, v):
        if not isinstance(v, float) or v < 0 or v > 1:
            raise ValueError("ice_gamma_fct must be a float between 0 and 1")
        return v

    @field_validator("depth_ice_fct")
    @classmethod
    def validate_depth_ice_fct(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("depth_ice_fct must be a positive float")
        return v

    @field_validator("h_ml0")
    @classmethod
    def validate_h_ml0(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("h_ml0 must be a positive float")
        return v

    @field_validator("salt_ice")
    @classmethod
    def validate_salt_ice(cls, v):
        if not isinstance(v, float) or v < 0:
            raise ValueError("salt_ice must be a non-negative float")
        return v

    @field_validator("salt_water")
    @classmethod
    def validate_salt_water(cls, v):
        if not isinstance(v, float) or v < 0:
            raise ValueError("salt_water must be a non-negative float")
        return v

    @field_validator("lead_closing")
    @classmethod
    def validate_lead_closing(cls, v):
        if not isinstance(v, float) or v <= 0:
            raise ValueError("lead_closing must be a positive float")
        return v

    @field_validator("saterm")
    @classmethod
    def validate_saterm(cls, v):
        if not isinstance(v, float) or v <= 0 or v > 1:
            raise ValueError("saterm must be a float between 0 and 1")
        return v

    @field_validator("albsn")
    @classmethod
    def validate_albsn(cls, v):
        if not isinstance(v, float) or v <= 0 or v > 1:
            raise ValueError("albsn must be a float between 0 and 1")
        return v

    @field_validator("albsnm")
    @classmethod
    def validate_albsnm(cls, v):
        if not isinstance(v, float) or v <= 0 or v > 1:
            raise ValueError("albsnm must be a float between 0 and 1")
        return v

    @field_validator("albi")
    @classmethod
    def validate_albi(cls, v):
        if not isinstance(v, float) or v <= 0 or v > 1:
            raise ValueError("albi must be a float between 0 and 1")
        return v

    @field_validator("albm")
    @classmethod
    def validate_albm(cls, v):
        if not isinstance(v, float) or v <= 0 or v > 1:
            raise ValueError("albm must be a float between 0 and 1")
        return v

    @model_validator(mode="after")
    def validate_rheol_steps(self):
        if self.ievp == 1 and self.evp_rheol_steps <= 0:
            raise ValueError("evp_rheol_steps must be positive when ievp is 1")
        return self

    @model_validator(mode="after")
    def validate_rheol_steps(self):
        if self.ievp == 2 and self.mevp_rheol_steps <= 0:
            raise ValueError("mevp_rheol_steps must be positive when ievp is 2")
        return self

    @model_validator(mode="after")
    def validate_cdwin0_usage(self):
        if self.ice_atmos_stress_form == 0 and self.cdwin0 <= 0:
            raise ValueError("cdwin0 must be positive when ice_atmos_stress_form is 0")
        return self

    @model_validator(mode="after")
    def validate_mevp_alpha1_usage(self):
        if self.mevp_coef == 0 and self.mevp_alpha1 <= 0:
            raise ValueError("mevp_alpha1 must be positive when mevp_coef is 0")
        return self

    @model_validator(mode="after")
    def validate_mevp_alpha2_usage(self):
        if self.mevp_coef == 0 and self.mevp_alpha2 <= 0:
            raise ValueError("mevp_alpha2 must be positive when mevp_coef is 0")
        return self

    @model_validator(mode="after")
    def validate_mevp_alpha3_usage(self):
        if self.mevp_coef == 1 and self.mevp_alpha3 <= 0:
            raise ValueError("mevp_alpha3 must be positive when mevp_coef is 1")
        return self

    @model_validator(mode="after")
    def validate_mevp_alpha4_usage(self):
        if self.mevp_coef == 1 and self.mevp_alpha4 <= 0:
            raise ValueError("mevp_alpha4 must be positive when mevp_coef is 1")
        return self

    @model_validator(mode="after")
    def validate_albsnm_relation(self):
        if self.albsnm > self.albsn:
            raise ValueError("albsnm must be less than or equal to albsn")
        return self

    @model_validator(mode="after")
    def validate_albi_relation(self):
        if self.albi > self.albsn:
            raise ValueError("albi must be less than or equal to albsn")
        return self

    @model_validator(mode="after")
    def validate_albm_relation(self):
        if self.albm > self.albi:
            raise ValueError("albm must be less than or equal to albi")
        return self


class Ice(NamelistBaseModel):
    ice_in: Optional[Ice_in] = Field(default_factory=Ice_in)



"""
Unit tests for SCHISM boundary functionality.

This module tests the boundary components and data handling for SCHISM.
"""

import pytest

from rompy.core.data import DataBlob
from rompy_schism.data import SCHISMDataBoundary

pytest.importorskip("rompy_schism")


class TestBoundaryHandling:
    """Tests for boundary handling functionality."""

    def test_create_boundary(self, grid2d):
        """Test creation of boundaries from a grid."""
        # Get boundary nodes
        boundary_nodes = (
            grid2d.get_boundary_nodes()
            if hasattr(grid2d, "get_boundary_nodes")
            else grid2d.ocean_boundary()[0]
        )

        # Verify we have boundary nodes
        assert len(boundary_nodes) > 0

    def test_data_boundary_creation(self, grid2d, hycom_bnd2d, tmp_path):
        """Test creating a data boundary object."""
        # Create a data boundary
        data_boundary = SCHISMDataBoundary(
            id="test_boundary",
            source=hycom_bnd2d.source,  # Use SourceFile directly from hycom_bnd2d
            variables=["surf_el"],
            sel_method="interp",
            time_buffer=[0, 1],
        )

        # Test that data boundary was created successfully
        assert data_boundary is not None
        assert data_boundary.variables == ["surf_el"]

        # Ensure boundary format is correct for the grid

    def test_ocean_boundary(self, grid2d, hycom_bnd2d):
        """Test ocean boundary data handling."""
        # Create ocean boundary data directly
        ocean_boundary = SCHISMDataBoundary(
            id="elev2D",
            source=hycom_bnd2d.source,  # Use SourceFile directly
            variables=["surf_el"],
        )

        assert ocean_boundary is not None
        assert ocean_boundary.id == "elev2D"

    def test_3d_boundary(self, grid3d, hycom_bnd_temp_3d, tmp_path):
        """Test 3D boundary data handling."""
        # Create 3D temperature boundary
        temp_boundary = SCHISMDataBoundary(
            id="TEM_3D",
            source=hycom_bnd_temp_3d.source,  # Use SourceFile directly
            variables=["water_temp"],
            sel_method="interp",
        )

        # Test that 3D boundary was created successfully
        assert temp_boundary is not None
        assert temp_boundary.variables == ["water_temp"]

        # Ensure boundary format is correct for the grid

    def test_boundary_validation(self):
        """Test validation in boundary objects."""
        # Test with missing required fields
        with pytest.raises(Exception):
            # No source provided
            SCHISMDataBoundary(
                id="validation_test",
                variables=["surf_el"],
            )

        with pytest.raises(Exception):
            # Empty variables list
            SCHISMDataBoundary(
                id="validation_test2",
                source=DataBlob(source="dummy"),
                variables=[],
            )

    def test_boundary_to_file(self, grid2d, hycom_bnd2d, tmp_path):
        """Test writing boundary data to files."""
        # Create a data boundary
        SCHISMDataBoundary(
            id="test_boundary_file",
            source=hycom_bnd2d.source,  # Use SourceFile directly
            variables=["surf_el"],
            sel_method="interp",
        )

        # Here we would test writing to a file, if that functionality is implemented
        # This is a placeholder for when that functionality is available


if __name__ == "__main__":
    pytest.main(["-v", __file__])



import warnings
from pathlib import Path
from typing import Literal, Optional

from pydantic import ConfigDict, Field, model_serializer, model_validator

from rompy.core.config import BaseConfig
from rompy.logging import get_logger

from .config_legacy import SchismCSIROConfig as _LegacySchismCSIROConfig

# Import plotting functions
from .config_plotting import plot_sflux_spatial, plot_sflux_timeseries
from .config_plotting_boundary import (
    plot_boundary_points,
    plot_boundary_profile,
    plot_boundary_timeseries,
)
from .config_plotting_tides import (
    plot_tidal_boundaries,
    plot_tidal_dataset,
    plot_tidal_rose,
    plot_tidal_stations,
)
from .data import SCHISMData
from .grid import SCHISMGrid
from .namelists import NML
from .namelists.param import Param

logger = get_logger(__name__)

HERE = Path(__file__).parent

SCHISM_TEMPLATE = str(Path(__file__).parent / "templates" / "schism")


class SCHISMConfig(BaseConfig):
    model_type: Literal["schism"] = Field(
        "schism", description="The model type for SCHISM."
    )
    grid: SCHISMGrid = Field(description="The model grid")
    data: Optional[SCHISMData] = Field(None, description="Model inputs")
    nml: Optional[NML] = Field(
        default_factory=lambda: NML(param=Param()), description="The namelist"
    )
    template: Optional[str] = Field(
        description="The path to the model template",
        default=SCHISM_TEMPLATE,
    )

    # add a validator that checks that nml.param.ihot is 1 if data.hotstart is not none
    @model_validator(mode="after")
    def check_hotstart(self):
        if (
            self.data is not None
            and hasattr(self.data, "hotstart")
            and getattr(self.data, "hotstart", None) is not None
            and self.nml is not None
            and self.nml.param is not None
            and self.nml.param.opt is not None
        ):
            self.nml.param.opt.ihot = 1
        return self

    @model_serializer
    def serialize_model(self, **kwargs):
        """Custom serializer to handle proper serialization of nested components."""
        from rompy_schism.grid import GR3Generator

        result = {}

        # Explicitly handle required fields
        result["model_type"] = self.model_type

        # Handle grid separately to process GR3Generator objects
        if self.grid is not None:
            grid_dict = {}
            for field_name in self.grid.model_fields:
                value = getattr(self.grid, field_name, None)

                # Special handling for GR3Generator objects
                if value is not None and isinstance(value, GR3Generator):
                    # For GR3Generator objects, extract just the value field
                    grid_dict[field_name] = value.value
                elif value is not None and not field_name.startswith("_"):
                    grid_dict[field_name] = value

            result["grid"] = grid_dict

        # Add optional fields that are not None
        if self.data is not None:
            result["data"] = self.data

        if self.nml is not None:
            result["nml"] = self.nml

        if self.template is not None:
            result["template"] = self.template

        return result

    # Enable arbitrary types and validation from instances in Pydantic v2
    model_config = ConfigDict(arbitrary_types_allowed=True, from_attributes=True)

    # Add data visualization methods
    # Atmospheric (sflux) plotting
    plot_sflux_spatial = plot_sflux_spatial
    plot_sflux_timeseries = plot_sflux_timeseries

    # Boundary data plotting
    plot_boundary_points = plot_boundary_points
    plot_boundary_timeseries = plot_boundary_timeseries
    plot_boundary_profile = plot_boundary_profile

    # Tidal data plotting
    plot_tidal_boundaries = plot_tidal_boundaries
    plot_tidal_stations = plot_tidal_stations
    plot_tidal_rose = plot_tidal_rose
    plot_tidal_dataset = plot_tidal_dataset

    def __call__(self, runtime) -> str:
        from rompy.formatting import ARROW, log_box

        # Grid generation section
        log_box(
            title="GENERATING GRID FILES",
            logger=logger,
            add_empty_line=False,
        )
        logger.info(f"{ARROW} Grid type: {type(self.grid).__name__}")

        # Log grid details
        if hasattr(self.grid, "hgrid") and self.grid.hgrid is not None:
            hgrid_source = str(getattr(self.grid.hgrid, "source", self.grid.hgrid))
            if len(hgrid_source) > 60:
                hgrid_source = "..." + hgrid_source[-57:]
            logger.info(f"{ARROW} Horizontal grid: {hgrid_source}")

        if hasattr(self.grid, "vgrid") and self.grid.vgrid is not None:
            vgrid_type = type(self.grid.vgrid).__name__
            logger.info(f"{ARROW} Vertical grid: {vgrid_type}")

        # Generate grid files
        self.grid.get(runtime.staging_dir)
        logger.info(f"{ARROW} Grid files generated successfully")

        # Data processing section
        if self.data is not None:
            log_box(
                title="PROCESSING INPUT DATA",
                logger=logger,
                add_empty_line=False,
            )

            # Log data components
            data_components = []
            if hasattr(self.data, "atmos") and self.data.atmos is not None:
                data_components.append("Atmospheric")
            if hasattr(self.data, "wave") and self.data.wave is not None:
                data_components.append("Wave")
            if (
                hasattr(self.data, "boundary_conditions")
                and self.data.boundary_conditions is not None
            ):
                data_components.append("Boundary Conditions")

            if data_components:
                logger.info(f"{ARROW} Components: {', '.join(data_components)}")

            # Process data
            data_results = self.data.get(
                destdir=runtime.staging_dir, grid=self.grid, time=runtime.period
            )

            if self.nml is not None:
                self.nml.update_data_sources(data_results)

            logger.info(f"{ARROW} Input data processed successfully")

        # Namelist configuration section
        if self.nml is not None:
            log_box(
                title="CONFIGURING NAMELISTS",
                logger=logger,
                add_empty_line=False,
            )

            # Log active modules
            active_modules = []
            if hasattr(self.nml, "param") and self.nml.param is not None:
                active_modules.append("Parameters")
            if hasattr(self.nml, "ice") and self.nml.ice is not None:
                active_modules.append("Ice")
            if hasattr(self.nml, "icm") and self.nml.icm is not None:
                active_modules.append("ICM")
            if hasattr(self.nml, "sediment") and self.nml.sediment is not None:
                active_modules.append("Sediment")
            if hasattr(self.nml, "wwminput") and self.nml.wwminput is not None:
                active_modules.append("Wave")

            if active_modules:
                logger.info(f"{ARROW} Active modules: {', '.join(active_modules)}")

            # Update times and write namelists
            self.nml.update_times(period=runtime.period)
            self.nml.write_nml(runtime.staging_dir)
            logger.info(f"{ARROW} Namelists configured successfully")

        return str(runtime.staging_dir)

    def _format_value(self, obj):
        """Custom formatter for SCHISMConfig values.

        This method provides special formatting for specific types used in
        SCHISMConfig such as grid, data, and namelist components.

        Args:
            obj: The object to format

        Returns:
            A formatted string or None to use default formatting
        """
        # Import specific types and formatting utilities
        from rompy.formatting import get_formatted_header_footer
        from rompy.logging import LoggingConfig

        # Get ASCII mode setting from LoggingConfig
        logging_config = LoggingConfig()
        USE_ASCII_ONLY = logging_config.use_ascii

        # Format SCHISMConfig (self-formatting)
        if isinstance(obj, SCHISMConfig):
            header, footer, bullet = get_formatted_header_footer(
                title="SCHISM MODEL CONFIGURATION", use_ascii=USE_ASCII_ONLY
            )

            lines = [header]

            # Add grid information
            if obj.grid is not None:
                grid_type = type(obj.grid).__name__
                lines.append(f"  {bullet} Grid: {grid_type}")

                # Try to get grid details
                if hasattr(obj.grid, "hgrid"):
                    hgrid = obj.grid.hgrid
                    hgrid_path = str(
                        getattr(hgrid, "uri", getattr(hgrid, "source", hgrid))
                    )
                    if len(hgrid_path) > 50:
                        hgrid_path = "..." + hgrid_path[-47:]
                    lines.append(f"      Horizontal grid: {hgrid_path}")

                if hasattr(obj.grid, "vgrid") and obj.grid.vgrid is not None:
                    vgrid_type = type(obj.grid.vgrid).__name__
                    lines.append(f"      Vertical grid: {vgrid_type}")

            # Add data information
            if obj.data is not None:
                data_type = type(obj.data).__name__
                lines.append(f"  {bullet} Data: {data_type}")

                # Count data components
                data_components = []
                if hasattr(obj.data, "atmos") and obj.data.atmos is not None:
                    data_components.append("Atmospheric")
                if hasattr(obj.data, "wave") and obj.data.wave is not None:
                    data_components.append("Wave")
                if (
                    hasattr(obj.data, "boundary_conditions")
                    and obj.data.boundary_conditions is not None
                ):
                    data_components.append("Boundary Conditions")

                if data_components:
                    lines.append(f"      Components: {', '.join(data_components)}")

            # Add namelist information
            if obj.nml is not None:
                nml_type = type(obj.nml).__name__
                lines.append(f"  {bullet} Namelist: {nml_type}")

                # Count active namelist components
                nml_components = []
                if hasattr(obj.nml, "param") and obj.nml.param is not None:
                    nml_components.append("Parameters")
                if hasattr(obj.nml, "ice") and obj.nml.ice is not None:
                    nml_components.append("Ice")
                if hasattr(obj.nml, "icm") and obj.nml.icm is not None:
                    nml_components.append("ICM")
                if hasattr(obj.nml, "sediment") and obj.nml.sediment is not None:
                    nml_components.append("Sediment")
                if hasattr(obj.nml, "wwminput") and obj.nml.wwminput is not None:
                    nml_components.append("Wave")

                if nml_components:
                    lines.append(f"      Active modules: {', '.join(nml_components)}")

            # Add template information
            if obj.template is not None:
                template_path = obj.template
                if len(template_path) > 50:
                    template_path = "..." + template_path[-47:]
                lines.append(f"  {bullet} Template: {template_path}")

            lines.append(footer)
            return "\n".join(lines)

        # Format SCHISMGrid
        from .grid import SCHISMGrid

        if isinstance(obj, SCHISMGrid):
            header, footer, bullet = get_formatted_header_footer(
                title="SCHISM GRID", use_ascii=USE_ASCII_ONLY
            )

            lines = [header]

            if hasattr(obj, "hgrid"):
                hgrid = obj.hgrid
                hgrid_path = str(getattr(hgrid, "uri", getattr(hgrid, "source", hgrid)))
                if len(hgrid_path) > 50:
                    hgrid_path = "..." + hgrid_path[-47:]
                lines.append(f"  {bullet} Horizontal Grid: {hgrid_path}")

            if hasattr(obj, "vgrid") and obj.vgrid is not None:
                vgrid_type = type(obj.vgrid).__name__
                lines.append(f"  {bullet} Vertical Grid: {vgrid_type}")

                # Try to get vertical grid details
                if hasattr(obj.vgrid, "nlayer") and hasattr(obj.vgrid, "nlayer"):
                    nlayer = getattr(obj.vgrid, "nlayer", None)
                    if nlayer is not None:
                        lines.append(f"      Layers: {nlayer}")

            lines.append(footer)
            return "\n".join(lines)

        # Format SCHISMData
        from .data import SCHISMData

        if isinstance(obj, SCHISMData):
            header, footer, bullet = get_formatted_header_footer(
                title="SCHISM DATA", use_ascii=USE_ASCII_ONLY
            )

            lines = [header]

            # Count and list data components
            components = {}
            if hasattr(obj, "atmos") and obj.atmos is not None:
                components["Atmospheric"] = type(obj.atmos).__name__
            if hasattr(obj, "wave") and obj.wave is not None:
                components["Wave"] = type(obj.wave).__name__
            if (
                hasattr(obj, "boundary_conditions")
                and obj.boundary_conditions is not None
            ):
                components["Boundary Conditions"] = type(
                    obj.boundary_conditions
                ).__name__

            for comp_name, comp_type in components.items():
                lines.append(f"  {bullet} {comp_name}: {comp_type}")

            if not components:
                lines.append(f"  {bullet} No data components configured")

            lines.append(footer)
            return "\n".join(lines)

        # Format NML
        from .namelists import NML

        if isinstance(obj, NML):
            header, footer, bullet = get_formatted_header_footer(
                title="SCHISM NAMELIST", use_ascii=USE_ASCII_ONLY
            )

            lines = [header]

            # List active namelist components
            components = {}
            if hasattr(obj, "param") and obj.param is not None:
                components["Parameters"] = type(obj.param).__name__
            if hasattr(obj, "ice") and obj.ice is not None:
                components["Ice"] = type(obj.ice).__name__
            if hasattr(obj, "icm") and obj.icm is not None:
                components["ICM"] = type(obj.icm).__name__
            if hasattr(obj, "sediment") and obj.sediment is not None:
                components["Sediment"] = type(obj.sediment).__name__
            if hasattr(obj, "wwminput") and obj.wwminput is not None:
                components["Wave"] = type(obj.wwminput).__name__
            if hasattr(obj, "cosine") and obj.cosine is not None:
                components["CoSiNE"] = type(obj.cosine).__name__
            if hasattr(obj, "mice") and obj.mice is not None:
                components["MICE"] = type(obj.mice).__name__

            for comp_name, comp_type in components.items():
                lines.append(f"  {bullet} {comp_name}: {comp_type}")

            if not components:
                lines.append(f"  {bullet} No modules configured")

            lines.append(footer)
            return "\n".join(lines)

        # Use the new formatting framework
        from rompy.formatting import format_value

        return format_value(obj)


class SchismCSIROConfig(_LegacySchismCSIROConfig):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "The SchismCSIROMigrationConfig class from config.py is deprecated. "
        )
        super().__init__(*args, **kwargs)
